# Generated by Django 4.0.2 on 2022-02-10 11:55

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('_event_provider', models.JSONField(blank=True, default=dict, verbose_name='Event Provider')),
                ('owner', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Owner')),
            ],
            options={
                'verbose_name': 'Profile',
                'verbose_name_plural': 'Profiles',
            },
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(help_text='The start date and time for this recurring event.', verbose_name='Start Date')),
                ('title', models.CharField(help_text='This will be included in the generated output and is capped at 50 characters.', max_length=50, verbose_name='Event Title')),
                ('category', models.CharField(choices=[('ANNUAL_ANNIVERSARY', 'Annual Anniversary'), ('HOLIDAY', 'Holiday')], default='ANNUAL_ANNIVERSARY', help_text='The category of this event.', max_length=18, verbose_name='Event Category')),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='events', to='calingen.profile', verbose_name='Profile')),
            ],
            options={
                'verbose_name': 'Event',
                'verbose_name_plural': 'Events',
                'unique_together': {('title', 'start')},
            },
        ),
    ]



# SPDX-License-Identifier: MIT

"""App-specific exceptions."""


class CalingenException(Exception):
    """Base class for all app-specific exceptions."""


class CallingenInterfaceException(CalingenException):
    """Base class for all exceptions of the ``interfaces`` module."""



# SPDX-License-Identifier: MIT

"""Implementation of :class:`~calingen.interfaces.plugin_api.LayoutProvider` that provides TeX-sources for a simple list of events."""

# local imports
from .simple_event_list import SimpleEventList  # noqa: F401

default_app_config = "calingen.contrib.layouts.simple_event_list.apps.CalingenLayoutSimpleEventListConfig"
"""The path to the app's default configuration class.

Consider this *legacy code*. See
:djangoapi:`Django's documentation<applications/#configuring-applications>` for
details.

Notes
-----
While the layout is provided as a "standalone Django app", it is in fact nothing
more than an implementation of
:class:`~calingen.interfaces.plugin_api.LayoutProvider`. The actual magic of
registering the layout with the main app is not done in this app's
:class:`~django.apps.AppConfig`, but in this very file by importing the
:class:`~calingen.contrib.layouts.simple_event_list.simple_event_list.SimpleEventList`
class.
"""



# SPDX-License-Identifier: MIT

"""Implementation of :class:`~calingen.interfaces.plugin_api.LayoutProvider` that provides different kinds of ruled paper.

This layout is HTML-based and intended to be used to generate single pages with
a given grid (*"Lineatur"*).

Warnings
--------
This layout does not include events!
"""

# local imports
from .lineatur import Lineatur  # noqa: F401

default_app_config = (
    "calingen.contrib.layouts.lineatur.apps.CalingenLayoutLineaturConfig"
)
"""The path to the app's default configuration class.

Consider this *legacy code*. See
:djangoapi:`Django's documentation<applications/#configuring-applications>` for
details.

Notes
-----
While the layout is provided as a "standalone Django app", it is in fact nothing
more than an implementation of
:class:`~calingen.interfaces.plugin_api.LayoutProvider`. The actual magic of
registering the layout with the main app is not done in this app's
:class:`~django.apps.AppConfig`, but in this very file by importing the
:class:`~calingen.contrib.layouts.lineatur.lineatur.Lineatur` class.
"""



# SPDX-License-Identifier: MIT

"""Provide the application configuration for Django."""

# Django imports
from django.apps import AppConfig


class CalingenLayoutLineaturConfig(AppConfig):
    """Application-specific configuration class, as required by Django.

    This sub-class of Django's `AppConfig` provides application-specific
    information to be used in Django's application registry (see
    :djangoapi:`applications/#configuring-applications`).
    """

    name = "calingen.contrib.layouts.lineatur"
    verbose_name = "CalInGen Layout: Lineatur"



# SPDX-License-Identifier: MIT

"""The actual implementation of the compiler."""


# Django imports
from django.http.response import HttpResponse

# app imports
from calingen.interfaces.plugin_api import CompilerProvider

SOURCE_TYPE_LOOKUP = {
    "plain": ("txt", "text/plain; charset=UTF-8"),
    "html": ("html", "text/html; charset=UTF-8"),
    "markdown": ("md", "text/markdown; charset=UTF-8"),
    "rst": ("rst", "text/x-rst"),
    "tex": ("tex", "application/x-tex"),
}
"""A :py:obj:`dict` that provides meta information for file downloads.

The :meth:`~calingen.contrib.compilers.download.compiler.DownloadCompiler.get_response`
uses this :py:obj:`dict` to determine the file extension and MIME type for the
download, based on the layout's ``layout_type`` attribute (see
:class:`calingen.interfaces.plugin_api.LayoutProvider`).
"""


class DownloadCompiler(CompilerProvider):
    """Do not compile sources, but provide them as file download."""

    title = "DownloadCompiler"

    @classmethod
    def get_response(cls, source, layout_type=None):  # noqa: D102
        if layout_type is not None:  # pragma: nocover
            try:
                file_extension = SOURCE_TYPE_LOOKUP[layout_type][0]
                content_type = SOURCE_TYPE_LOOKUP[layout_type][1]
            except KeyError:
                file_extension = SOURCE_TYPE_LOOKUP["plain"][0]
                content_type = SOURCE_TYPE_LOOKUP["plain"][1]
        else:  # pragma: nocover
            file_extension = SOURCE_TYPE_LOOKUP["plain"][0]
            content_type = SOURCE_TYPE_LOOKUP["plain"][1]

        return HttpResponse(
            source,
            headers={
                "Content-Type": content_type,
                "Content-Disposition": 'attachment; filename="calingen_generated_source.{}"'.format(
                    file_extension
                ),
            },
        )  # pragma: nocover



# SPDX-License-Identifier: MIT

""":class:`~calingen.interfaces.plugin_api.LayoutProvider` implementation for various ruled paper outputs.

More or less all of this module's source code is utility code, that is tied
together in :class:`~calingen.contrib.layouts.lineatur.lineatur.LineaturForm`
and actually used in the layout's templates.

This may seem like a lot of boilerplate (and it is, in fact), but it
demonstrates an easy way to actually implement
:ref:`the LayoutProvider's ability to provide a layout specific configuration <calingen-dev-doc-plugins-layoutprovider-label>`
as implemented by
:attr:`calingen.interfaces.plugin_api.LayoutProvider.configuration_form`.
"""

# Django imports
from django.db.models import TextChoices
from django.forms import CharField, ChoiceField, FloatField, IntegerField
from django.utils.translation import gettext_lazy as _

# app imports
from calingen.forms.generation import LayoutConfigurationForm
from calingen.interfaces.plugin_api import LayoutProvider


class PaperSize(TextChoices):
    """Provide choices for the paper sizes."""

    A0 = "paperA0", _("A0 (841mm x 1189mm)")
    A1 = "paperA1", _("A1 (594mm x 841mm)")
    A2 = "paperA2", _("A2 (420mm x 594mm)")
    A3 = "paperA3", _("A3 (297mm x 420mm)")
    A4 = "paperA4", _("A4 (210mm x 297mm)")
    A5 = "paperA5", _("A5 (148mm x 210mm)")
    A6 = "paperA6", _("A6 (105mm x 148mm)")
    A7 = "paperA7", _("A7 (74mm x 105mm)")
    B0 = "paperB0", _("B0 (1000mm x 1414mm)")
    B1 = "paperB1", _("B1 (707mm x 1000mm)")
    B2 = "paperB2", _("B2 (500mm x 707mm)")
    B3 = "paperB3", _("B3 (353mm x 500mm)")
    B4 = "paperB4", _("B4 (250mm x 353mm)")
    B5 = "paperB5", _("B5 (176mm x 250mm)")
    B6 = "paperB6", _("B6 (125mm x 176mm)")
    B7 = "paperB7", _("B7 (88mm x 125mm)")
    LETTER = "paperLetter", _("Letter (8.5in x 11.0in)")
    GOV_LETTER = "paperGovLetter", _("Government Letter (8.0in x 10.5in)")
    LEGAL = "paperLegal", _("Government Legal (8.5in x 14.0in)")
    LEDGER = "paperLedger", _("Ledger (17.0in x 11.0in)")
    TABLOID = "paperTabloid", _("Tabloid (11.0in x 17.0in)")


class LengthUnits(TextChoices):
    """Provide CSS absolute length units.

    Taken from here:
    https://developer.mozilla.org/en-US/docs/Learn/CSS/Building_blocks/Values_and_units#lengths
    """

    CM = "cm", _("Centimeters")
    MM = "mm", _("Millimeters")
    IN = "in", _("Inches")
    PT = "pt", _("Points (1/72th of 1in)")
    PX = "px", _("Pixels (1/96th of 1in)")


class LineaturTypes(TextChoices):
    """Provide different types of grids."""

    BLANK = "blank", _("blank")
    DOTTED = "dotted", _("dotted")
    LINED = "lined", _("lined")
    SQUARED = "squared", _("squared")


class CaptionPositionChoices(TextChoices):
    """Where to put the caption."""

    LEFT = "left", _("left")
    CENTER = "center", _("center")
    RIGHT = "right", _("right")


class ColorChoices(TextChoices):
    """Pre-defined colors to use in the generated output."""

    BLACK = "rgb(0,0,0)", _("black")
    DARK_GREY = "rgb(64,64,64)", _("dark grey")
    LIGHT_GREY = "rgb(128,128,128)", _("light grey")
    RED = "rgb(255,0,0)", _("red")
    GREEN = "rgb(0,255,0)", _("green")
    BLUE = "rgb(0,0,255)", _("blue")


class LineaturForm(LayoutConfigurationForm):
    """Configuration form for the Lineatur layout.

    This is actually a subclass of
    :class:`calingen.forms.generation.LayoutConfigurationForm` and exposes some
    options to adjust the generated page, e.g. the format of the paper, margins
    and the actual grid.
    """

    template_name = "lineatur/config_form.html"

    caption = CharField(
        label=_("Caption"),
        help_text=_("An optional caption, placed in the specified position"),
        max_length=150,
        required=False,
        empty_value="",
    )

    caption_position = ChoiceField(
        label=_("Position of the Caption"),
        help_text=_("Where to put the caption"),
        choices=CaptionPositionChoices.choices,
        required=True,
        initial=CaptionPositionChoices.RIGHT,
    )

    caption_size = IntegerField(
        label=_("Size of the Caption"),
        help_text=_("Size in pt of the caption"),
        required=True,
        initial=16,
        min_value=1,
    )

    caption_font = CharField(
        label=_("Font Family of the Caption"),
        help_text=_(
            "Specify the font to use; will be passed to a CSS font-family attribute; "
            "You may specify system fonts; default: sans-serif."
        ),
        required="False",
        empty_value="sans-serif",
    )

    caption_color = ChoiceField(
        label=_("Caption Text Color"),
        help_text=_("The color to be used for the caption."),
        choices=ColorChoices.choices,
        required=True,
        initial=ColorChoices.LIGHT_GREY,
    )

    paper_size = ChoiceField(
        label=_("Paper Size"),
        help_text=_("The paper size to be used"),
        choices=PaperSize.choices,
        required=True,
        initial=PaperSize.A5,
    )

    lineatur_type = ChoiceField(
        label=_("Type of Grid to generate"),
        help_text=_("Determine which grid is generated"),
        choices=LineaturTypes.choices,
        required=True,
        initial=LineaturTypes.BLANK,
    )

    length_unit = ChoiceField(
        label=_("Unit of lengths"),
        help_text=_("Determine the unit for specified lengths"),
        choices=LengthUnits.choices,
        required=True,
        initial=LengthUnits.CM,
    )

    lineatur_spacing_x = FloatField(
        label=_("Grid horizontal spacing"),
        help_text=_("The horizontal spacing of the grid. See Unit of lengths."),
        required=True,
        initial=0.5,
    )

    lineatur_spacing_y = FloatField(
        label=_("Grid vertical spacing"),
        help_text=_("The vertical spacing of the grid. See Unit of lengths."),
        required=True,
        initial=0.5,
    )

    lineatur_color = ChoiceField(
        label=_("Grid Color"),
        help_text=_("The color to render the selected grid"),
        choices=ColorChoices.choices,
        required=True,
        initial=ColorChoices.LIGHT_GREY,
    )

    lineatur_stroke_width = IntegerField(
        label=_("Width of the grid strokes"),
        help_text=_("The width of the grid strokes specified in pixels"),
        required=True,
        initial=1,
        min_value=0,
    )

    page_margin_top = FloatField(
        label=_("Margin (top)"),
        help_text=_(
            "Space between the page's and the content's top edges. "
            "See Unit of lengths."
        ),
        required=True,
        initial=0,
    )

    page_margin_right = FloatField(
        label=_("Margin (right)"),
        help_text=_(
            "Space between the page's and the content's right edges. "
            "See Unit of lengths."
        ),
        required=True,
        initial=0,
    )

    page_margin_bottom = FloatField(
        label=_("Margin (bottom)"),
        help_text=_(
            "Space between the page's and the content's bottom edges. "
            "See Unit of lengths."
        ),
        required=True,
        initial=0,
    )

    page_margin_left = FloatField(
        label=_("Margin (left)"),
        help_text=_(
            "Space between the page's and the content's left edges. "
            "See Unit of lengths."
        ),
        required=True,
        initial=0,
    )


class Lineatur(LayoutProvider):
    """Provide different kinds of ruled paper.

    The term *lineatur* can be translated (loosely) to *ruled paper*. It
    describes the lines or grids on paper, usually used for notes.
    """

    name = "Lineatur"
    paper_size = "various"
    orientation = "portrait"
    layout_type = "html"
    configuration_form = LineaturForm
    _template = "lineatur/layout.html"



# SPDX-License-Identifier: MIT

"""Implementations of :class:`django.forms.Form` with generic functionality."""

# Django imports
from django.forms import Form


class RequestEnabledForm(Form):
    """Form class with the ``request`` object available.

    Parameters
    ----------
    request
        Beside ``*args`` and ``**kwargs``, which make this implementation
        compatible with Django's :class:`~django.forms.Form`, it accepts the
        keyword argument ``request`` and stores it within the object.

    Notes
    -----
    This form implementation is intended to be used together with
    :class:`calingen.views.generic.RequestEnabledFormView`.

    In fact, an actual
    form will subclass this form and use a view that is a subclass of
    :class:`~calingen.views.generic.RequestEnabledFormView`.
    """

    def __init__(self, *args, request=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request



# SPDX-License-Identifier: MIT

""":class:`~calingen.interfaces.plugin_api.LayoutProvider` implementation for a year's calendar.

Warnings
--------
This layout is included in **django-calingen**'s CI test setup, mainly to verify
that the :func:`TeX escaping <calingen.templatetags.calingen_escape.escape_tex>`
is working.

This may be object to future changes.
"""

# Python imports
from datetime import date, timedelta

# Django imports
from django.template.defaultfilters import date as _date

# app imports
from calingen.constants import EventCategory
from calingen.interfaces.plugin_api import LayoutProvider


class CalendarDay:
    """Layout-specific class to manage calendar days.

    Parameters
    ----------
    date : datetime.date
        The Python representation of the date.
    date_entries : list
        A list of ``entries`` for the given ``date`` (will be a subset of the
        provided :class:`calingen.interfaces.data_exchange.CalendarEntryList`
        and each entry is expected to be an instance of
        :class:`calingen.interfaces.data_exchange.CalendarEntry`).
    """

    def __init__(self, date, date_entries):
        self.date = date

        # filter date_entries into categories
        self.holidays = [
            x.title for x in date_entries if x.category == EventCategory.HOLIDAY
        ]
        self.annuals = [
            x.title
            for x in date_entries
            if x.category == EventCategory.ANNUAL_ANNIVERSARY
        ]


class CalendarWeek:
    """Layout-specific class to manage calendar weeks."""

    def __init__(self):
        self.calendarweek = None
        self.month_string = None
        self.days = []

        self.populate_week = self._populate_week
        self.check_turnover = self._check_turnover

    def add_day(self, day):
        """Add a layout-specific day object to this week.

        This method is the actual public entry point of this class, allowing
        ``day`` objects to be added to the instance.

        Parameters
        ----------
        day : calingen.contrib.layouts.year_by_week.year_by_week.CalendarDay
            A layout-specific representation of a single ``day``, including its
            actual ``date`` (:py:obj:`datetime.date`) and instances of
            :class:`calingen.interfaces.data_exchange.CalendarEntry` associated
            with that ``date``.
        """
        self.populate_week(day.date)

        self.days.append(day)

        self.check_turnover(day.date)

    @property
    def is_empty(self):
        """Flag to indicate, if this object already contains days."""
        return not self.days

    def _check_turnover(self, date_of_day):
        """Update the ``month_string``.

        While adding days to this object, a *turnover* may happen, meaning that
        during that week a new month starts.

        This method updates the ``month_string`` attribute accordingly.

        As there can only be one *turnover* during any given week,
        :attr:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.check_turnover`
        is substituted with
        :meth:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek._noop`
        during its execution.

        Warnings
        --------
        This is a private method, other methods should use
        :attr:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.check_turnover`
        to ensure performance.

        Parameters
        ----------
        date_of_day : datetime.date
            A Python date representation.
        """
        reference = self.days[0].date

        # no turnover, just return
        if reference.month == date_of_day.month:
            return

        if reference.year != date_of_day.year:
            # Translation of month names
            # https://stackoverflow.com/a/6991918
            self.month_string = (
                "{month_ref} {year_ref}/{month_this} {year_this}".format(
                    month_ref=_date(reference, "F"),
                    year_ref=reference.year,
                    month_this=_date(date_of_day, "F"),
                    year_this=date_of_day.year,
                )
            )
        else:
            # Translation of month names
            # https://stackoverflow.com/a/6991918
            self.month_string = "{month_ref}/{month_this}".format(
                month_ref=_date(reference, "F"),
                month_this=_date(date_of_day, "F"),
            )

        # Turnovers may happen only once per week, so provide noop() method
        # after first usage.
        self.check_turnover = self._noop

    def _populate_week(self, date_of_day):
        """Set the object's meta data, e.g. number of the calendar week.

        This is a one-time operation that is performed on the first call to
        :meth:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.add_day`.
        During that call,
        :attr:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.populate_week`
        is substituted with
        :meth:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek._noop`.

        Warnings
        --------
        This is a private method, other methods should use
        :attr:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.populate_week`
        to ensure performance.

        Parameters
        ----------
        date_of_day : datetime.date
            A Python date representation.
        """
        # isocalendar() returns a tuple of (year, week, weekday)
        self.calendarweek = date_of_day.isocalendar()[1]

        # as this method is only executed once (per week), just apply the month
        # of the given day
        # Translation of month names
        # https://stackoverflow.com/a/6991918
        self.month_string = _date(date_of_day, "F")

        # Population of the week object is a one-time operation, so provide the
        # noop() method after the first usage.
        self.populate_week = self._noop

    def _noop(self, *args, **kwargs):
        """Just a generic no op method.

        This method is used to substitute
        :attr:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.populate_week`
        and
        :attr:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.check_turnover`
        during processing.
        """
        pass

    def __repr__(self):  # noqa: D105
        return "<CalendarWeek [{}] {}>".format(
            self.calendarweek, self.days[0].date
        )  # pragma: nocover

    def __str__(self):  # noqa: D105
        return "[{}]".format(self.calendarweek)  # pragma: nocover


class YearByWeek(LayoutProvider):
    """The actual implementation of the layout.

    :meth:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek.prepare_context`
    is used to create a year's full representation, grouping the days by week
    and matching the ``entries`` to the given day.

    Warnings
    --------
    The provided templates create a document targeted at German users. You may
    override the templates to (fully) support other languages.

    Notes
    -----
    To customize the generated TeX-sources, the following templates may be
    overridden:

    - ``year_by_week/tex/base.tex``: Speaking in TeX-terms: the preamble of
      the document, including package definitions.
    - ``year_by_week/tex/year_by_week.tex``: Speaking in TeX-terms:
      the document's body.
    """

    name = "Year by Week"
    paper_size = "a5"
    orientation = "portrait"
    layout_type = "tex"
    _template = "year_by_week/tex/year_by_week.tex"

    @classmethod
    def prepare_context(cls, context):
        """Create a full year's representation and return it as ``weeklist``."""
        # values from the context for processing
        target_year = context.get("target_year")
        entries = context.get("entries", [])

        # The first day of the calendar is the monday before [YEAR]-01-01
        # date().weekday() = 0 for Mondays, 6 for Sundays
        first_yearday = date(target_year, 1, 1) - timedelta(
            days=date(target_year, 1, 1).weekday()
        )
        # The last day of the calendar is the sunday after [YEAR]-12-31
        # date().weekday() = 0 for Mondays, 6 for Sundays
        # Note: This may seem to be "off by one", but this enables the while-loop
        # to match against "<" instead of "<=".
        last_yearday = date(target_year, 12, 31) + timedelta(
            days=(7 - date(target_year, 12, 31).weekday())
        )

        # prepare result
        weeklist = []
        this_week = CalendarWeek()
        # initialize the invariant
        day = first_yearday
        while day < last_yearday:
            if day.weekday() == 0:
                if not this_week.is_empty:
                    weeklist.append(this_week)
                    this_week = CalendarWeek()

            # This may look rather complex but is nothing more than filtering
            # the "entries" by day and feeding a list of entries into the
            # constructor of the CalendarDay class.
            this_day = CalendarDay(
                day,
                [
                    x
                    for x in entries
                    if x.timestamp.date().day == day.day
                    and x.timestamp.date().month == day.month
                ],
            )
            this_week.add_day(this_day)

            # increment the invariant
            day = day + timedelta(days=1)
        # append the final week to the result list
        if not this_week.is_empty:
            weeklist.append(this_week)

        context["weeklist"] = weeklist

        return context



# SPDX-License-Identifier: MIT

"""Implementation of :class:`~calingen.interfaces.plugin_api.LayoutProvider` that provides TeX-sources for a full year's calendar, presented by calendar weeks.

Warnings
--------
This layout does only include events of the types
:attr:`~calingen.constants.EventCategory.ANNUAL_ANNIVERSARY` and
:attr:`~calingen.constants.EventCategory.HOLIDAY`.
"""

# local imports
from .year_by_week import YearByWeek  # noqa: F401

default_app_config = (
    "calingen.contrib.layouts.year_by_week.apps.CalingenLayoutYearByWeekConfig"
)
"""The path to the app's default configuration class.

Consider this *legacy code*. See
:djangoapi:`Django's documentation<applications/#configuring-applications>` for
details.

Notes
-----
While the layout is provided as a "standalone Django app", it is in fact nothing
more than an implementation of
:class:`~calingen.interfaces.plugin_api.LayoutProvider`. The actual magic of
registering the layout with the main app is not done in this app's
:class:`~django.apps.AppConfig`, but in this very file by importing the
:class:`~calingen.contrib.layouts.year_by_week.year_by_week.YearByWeek`
class.
"""



# SPDX-License-Identifier: MIT

"""Calingen's implementations of :class:`django.db.models.Model`.

In the context of a Django application, ``models`` define the app's core data,
that is persisted in a database layer (abstracted by DJango's ORM).

This package also includes model-related implementations of
:class:`django.db.models.QuerySet`, :class:`django.db.models.Manager` and
:class:`django.forms.ModelForm`.

Notes
-----
It's kind of recommended to have *fat models* in a Django application, that
handle most of the app's business logic.

However, *fat models* might lead to models, that are *too tightly* connected to
each other.

Calingen's models aim to present and manage the app's data and enable working
with that data.
"""

# app imports
from calingen.models.event import Event  # noqa: F401
from calingen.models.profile import Profile  # noqa: F401



# SPDX-License-Identifier: MIT

"""App-specific URL configuration."""

# Python imports
import datetime

# Django imports
from django.urls import path

# app imports
from calingen.views import base, event, generation, profile, web

app_name = "calingen"
"""Define an application namespace for reversing URLs.

See :djangodoc:`URL namespaces <topics/http/urls/#url-namespaces>`.
"""

urlpatterns = [
    path("", base.homepage, name="homepage"),
    path("event/", event.EventListView.as_view(), name="event-list"),
    path("event/<int:event_id>/", event.EventDetailView.as_view(), name="event-detail"),
    path(
        "event/<int:event_id>/update/",
        event.EventUpdateView.as_view(),
        name="event-update",
    ),
    path(
        "event/<int:event_id>/delete/",
        event.EventDeleteView.as_view(),
        name="event-delete",
    ),
    path("event/add/", event.EventCreateView.as_view(), name="event-add"),
    path("<int:profile_id>/", profile.ProfileDetailView.as_view(), name="profile"),
    path("profile/add/", profile.ProfileCreateView.as_view(), name="profile-add"),
    path(
        "<int:profile_id>/delete/",
        profile.ProfileDeleteView.as_view(),
        name="profile-delete",
    ),
    path(
        "<int:profile_id>/update/",
        profile.ProfileUpdateView.as_view(),
        name="profile-update",
    ),
    path(
        "<int:profile_id>/events/",
        web.CalendarEntryListView.as_view(),
        kwargs={"target_year": datetime.datetime.now().year},
        name="calendar-entry-list-year",
    ),
    path(
        "<int:profile_id>/events/<int:target_year>/",
        web.CalendarEntryListView.as_view(),
        name="calendar-entry-list-year",
    ),
    path(
        "generate/select-layout/",
        generation.LayoutSelectionView.as_view(),
        name="layout-selection",
    ),
    path(
        "generate/configure-layout/",
        generation.LayoutConfigurationView.as_view(),
        name="layout-configuration",
    ),
    path(
        "generate/compilation/",
        generation.CompilerView.as_view(),
        name="compilation",
    ),
]
"""The actual patterns, matching URLs to views.

- ``""``: :func:`calingen.views.base.homepage` - ``"homepage"``

  The app's splash page. Will redirect to the user's profile (see
  ``"<profile_id>/"``) or to the profile creation page (see ``"profile/add/"``).
- ``"event/"``: :class:`calingen.views.event.EventListView` - ``"event-list"``

  The list of :class:`calingen.models.event.Event` instances for the requesting
  user.
- ``"event/<event_id>/"``: :class:`calingen.views.event.EventDetailView` - ``"event-detail"``

  The detailed view of an :class:`calingen.models.event.Event` instance.
- ``"event/add/"``: :class:`calingen.views.event.EventCreateView` - ``"event-add"``

  The view to create a new instance of :class:`calingen.models.event.Event`.
- ``"event/<event_id>/update/"``: :class:`calingen.views.event.EventUpdateView` - ``"event-update"``

  The view to modify / update an instance of :class:`calingen.models.event.Event`.
- ``"event/<event_id>/delete/"``: :class:`calingen.views.event.EventDeleteView` - ``"event-delete"``

  The view to delete an instance of :class:`calingen.models.event.Event`.
- ``"<profile_id>/"``: :class:`calingen.views.profile.ProfileDetailView` - ``"profile"``

  A view that summarizes the user's profile.
- ``"profile/add/"``: :class:`calingen.views.profile.ProfileCreateView` - ``"profile-add"``

  The view to create an instance of :class:`calingen.models.profile.Profile`.
- ``"<profile_id>/update/"``: :class:`calignen.views.profile.ProfileUpdateView` - ``"profile-update"``

  The view to modify / update instances of :class:`calingen.models.profile.Profile`.
- ``"<profile_id>/delete/"``: :class:`calingen.views.profile.ProfileDeleteView` - ``"profile-delete"``

  The view to delete an instance of :class:`calingen.models.profile.Profile`.
- ``"<profile_id>/events/"``: :class:`calingen.views.web.CalenderEntryListView` - ``"calendar-entry-list-year"``

  This view provides a list of a user's events. Those events are determined from
  external event providers (as configured in the user's profile) and the
  instances of :class:`calingen.models.event.Event` associated with the user.

  This view automatically determines events for the *current year*.

- ``"<profile_id>/events/<target_year>"``: :class:`calingen.views.web.CalendarEntryListView` -
  ``"calendar-entry-list-year"``

  See ``"<profile_id>/events/"``. The only difference is, that the
  ``target_year`` is provided manually through the URL.
- ``"generate/select-layout/"``: :class:`calingen.views.generation.LayoutSelectionView` - ``"layout-selection"``

  This view is the starting point of the actual *generation process*.
- ``"generate/configure-layout/"``: :class:`calingen.views.generation.LayoutConfigurationView` -
  ``"layout-configuration"``

  This view is automatically accessed during the *generation process*, if the
  selected layout requires configuration.
- ``"generate/compilation/"``: :class:`calingen.views.generation.CompilerView` - ``"compilation"``

  This view is automatically accessed during the *generation process* and
  presents the compilation result, depending on the determined compiler.
"""



# SPDX-License-Identifier: MIT

"""Basic views that are logically not connected to other components."""

# Django imports
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect

# app imports
from calingen.models.profile import Profile


@login_required
def homepage(request):
    """Provide the splash view for the app.

    This is the app's main entry point, redirecting the requesting user to his
    :class:`~calingen.models.profile.Profile` (provided by
    :class:`~calingen.views.profile.ProfileDetailView`) or to the
    :class:`~calingen.views.profile.ProfileCreateView` to create one.
    """
    # Every user has exactly one (calingen) Profile (One to One relation)
    # So, this fetches the Profile of the user or is None, if the user doesn't
    # have a Profile associated (yet)
    profile = Profile.calingen_manager.get_profile(request.user)

    # If there is no Profile, redirect to the view to add one
    if profile is None:
        return redirect("calingen:profile-add")

    # redirect to the user's Profile
    return redirect("calingen:profile", profile_id=profile.id)



# SPDX-License-Identifier: MIT

""":class:`~calingen.interfaces.plugin_api.LayoutProvider` implementation for a simple event list.

Warnings
--------
This layout is included in **django-calingen**'s CI test setup, mainly to verify
that the :func:`TeX escaping <calingen.templatetags.calingen_escape.escape_tex>`
is working.

This may be object to future changes.
"""

# app imports
from calingen.interfaces.plugin_api import LayoutProvider


class SimpleEventList(LayoutProvider):
    """The actual implementation of the layout.

    Nothing fancy here,
    :meth:`~calingen.contrib.layouts.simple_event_list.simple_event_list.SimpleEventList.prepare_context`
    is just used to group the provided
    :class:`calingen.interfaces.data_exchange.CalendarEntryList` by month.

    Warnings
    --------
    The provided templates create a document targeted at German users. You may
    override the templates to (fully) support other languages.

    Notes
    -----
    To customize the generated TeX-sources, the following templates may be
    overridden:

    - ``simple_event_list/tex/base.tex``: Speaking in TeX-terms: the preamble of
      the document, including package definitions.
    - ``simple_event_list/tex/simple_event_list.tex``: Speaking in TeX-terms:
      the document's body.
    - ``simple_event_list/tex/single_entry_line.tex``: The actual
      TeX-representation of a
      :class:`calingen.interfaces.data_exchange.CalendarEntry`.
    """

    name = "Simple Event List"
    paper_size = "a4"
    orientation = "portrait"
    layout_type = "tex"
    _template = "simple_event_list/tex/simple_event_list.tex"

    @classmethod
    def prepare_context(cls, context):
        """Pre-process the ``entries`` to group them by month."""
        entries = context.pop("entries", [])

        # put each month's entries in a dedicated list
        processed_entries = []
        try:
            loop_month = entries[0].timestamp.month
        except IndexError:
            loop_month = "No Entries"

        month_list = []
        for entry in entries:
            if entry.timestamp.month != loop_month:
                processed_entries.append(month_list)
                month_list = []
                loop_month = entry.timestamp.month
            month_list.append(entry)
        if month_list:
            processed_entries.append(month_list)

        context["entries"] = processed_entries
        return context



# SPDX-License-Identifier: MIT

"""App-specific ``Event`` model.

Beside the actual :class:`calingen.models.event.Event` model, this module
contains the related implementations of :class:`django.db.models.QuerySet`,
:class:`django.db.models.Manager` and :class:`django.forms.ModelForm`.
"""

# Python imports
import datetime

# Django imports
from django import forms
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

# app imports
from calingen.constants import EventCategory
from calingen.exceptions import CalingenException
from calingen.forms.fields import SplitDateTimeOptionalField
from calingen.interfaces.data_exchange import (
    SOURCE_INTERNAL,
    CalendarEntry,
    CalendarEntryList,
)
from calingen.models.profile import Profile
from calingen.models.queryset import CalingenQuerySet


class EventModelException(CalingenException):
    """Base class for all exceptions related to the :class:`~calingen.models.event.Event` model."""


class EventQuerySet(CalingenQuerySet):
    """App-specific implementation of :class:`django.db.models.QuerySet`.

    Notes
    -----
    This :class:`~django.db.models.QuerySet` implementation provides
    app-specific augmentations.

    The provided methods augment/extend the retrieved
    :class:`calingen.models.event.Event` instances by annotating them with
    additional information.
    """

    def default(self):
        """Return a :class:`~django.db.models.QuerySet` with annotations.

        Returns
        -------
        :class:`~django.db.models.QuerySet`
            The annotated queryset.
        """
        return self  # pragma: nocover

    def filter_by_user(self, user):
        """Filter the result set by the objects' :attr:`owners <calingen.models.profile.Profile.owner>`.

        Parameters
        ----------
        user :
            An instance of the project's user model, as specified by
            :setting:`AUTH_USER_MODEL`.

        Returns
        -------
        :class:`django.db.models.QuerySet`
            The filtered queryset.

        Notes
        -----
        Effectively, this method is used to ensure, that any user may only
        access objects, which are owned by him. This is the app's way of
        ensuring `row-level permissions`, because only owners are allowed to
        view (and modify) their events.
        """
        return self.filter(profile__owner=user)


class EventManager(models.Manager):
    """App-/model-specific implementation of :class:`django.db.models.Manager`.

    Notes
    -----
    This :class:`~django.db.models.Manager` implementation is used as an
    additional manager of :class:`~calingen.models.event.Event` (see
    :attr:`calingen.models.event.Event.calingen_manager`).

    This implementation inherits its functionality from
    :class:`django.db.models.Manager` and provides identical funtionality.
    Furthermore, it augments the retrieved objects with additional attributes,
    using the custom :class:`~django.db.models.QuerySet` implementation
    :class:`~calingen.models.event.EventQuerySet`.
    """

    def get_calendar_entry_list(self, user=None, year=None):
        """Return all instances as :class:`~calingen.interfaces.data_exchange.CalendarEntryList`.

        Parameters
        ----------
        user :
            The summary is provided for an actual user, filtered by
            :attr:`calingen.models.event.Event.owner`.

            Most likely you will want to pass ``request.user`` into the method.

        Returns
        -------
        :class:`~calingen.interfaces.data_exchange.CalendarEntryList`
            All :class:`~calingen.models.event.Event` instances of ``user``,
            converted into a ``CalendarEntryList``.
        """
        result = CalendarEntryList()
        for event in self.get_user_events_qs(user).iterator():
            result.merge(event.resolve(year))

        return result

    def get_queryset(self):
        """Use the app-/model-specific :class:`~calingen.models.event.EventQuerySet` by default.

        Returns
        -------
        :class:`django.models.db.QuerySet`
            This queryset is provided by
            :class:`calingen.models.event.EventQuerySet` and applies its
            :meth:`~calingen.models.event.EventQuerySet.default` method. The
            retrieved objects will be annotated with additional attributes.
        """
        return EventQuerySet(self.model, using=self._db).default()  # pragma: nocover

    def get_user_events_qs(self, user=None):
        """Provide a ``QuerySet`` containing all events of a given ``user``.

        Parameters
        ----------
        user :
            The summary is provided for an actual user, filtered by
            :attr:`calingen.models.event.Event.owner`.

            Most likely you will want to pass ``request.user`` into the method.

        Returns
        -------
        :class:`django.models.db.QuerySet`
            This queryset is provided by
            :class:`calingen.models.event.EventQuerySet` and applies its
            :meth:`~calingen.models.event.EventQuerySet.filter_by_user` method.
        """
        if user is None:
            raise EventModelException(
                "This method may only be called with a user object"
            )
        return self.get_queryset().filter_by_user(user)

    def summary(self, user=None):
        """Provide a user-specific summary of :class:`~calingen.models.event.Event` instances.

        Parameters
        ----------
        user :
            The summary is provided for an actual user, filtered by
            :attr:`calingen.models.event.Event.owner`.

            Most likely you will want to pass ``request.user`` into the method.

        Returns
        -------
        int
            As of now, the method only returns the number of instances.
        """
        return self.get_user_events_qs(user).count()


class Event(models.Model):
    """Represents one event in a user's calendar.

    Warnings
    --------
    The class documentation only includes code that is actually shipped by the
    `calingen` app. Inherited attributes/methods (provided by Django's
    :class:`~django.db.models.Model`) are not documented here.
    """

    objects = models.Manager()
    """The model's default manager.

    The default manager is set to :class:`django.db.models.Manager`, which is
    the default value. In order to add the custom :attr:`calingen_manager` as
    an *additional* manager, the default manager has to be provided explicitly
    (see :djangodoc:`topics/db/managers/#default-managers`).
    """

    calingen_manager = EventManager()
    """App-/model-specific manager, that provides additional functionality.

    This manager is set to
    :class:`calingen.models.event.EventManager`. Its implementation provides
    augmentations of `Event` objects, by annotating them on database level.
    This will reduce the number of required database queries, if attributes of
    the object are accessed.

    The manager has to be used explicitly.
    """

    profile = models.ForeignKey(
        Profile,
        on_delete=models.CASCADE,
        related_name="events",
        verbose_name=Profile._meta.verbose_name,
    )
    """Reference to a :class:`~calingen.models.profile.Profile` object.

    Notes
    -----
    This is implemented as a :class:`~django.db.models.ForeignKey` with
    ``on_delete=CASCADE``, meaning: if the referenced
    :class:`~calingen.models.profile.Profile` object is deleted,
    all referencing ``Event`` objects are discarded aswell.

    The backwards relation (see
    :attr:`ForeignKey.related_name<django.db.models.ForeignKey.related_name>`)
    is named ``"events"``.
    """

    start = models.DateTimeField(
        help_text=_("The start date and time for this recurring event."),
        verbose_name=_("Start Date"),
    )
    """Date and time of a recurring event (:py:obj:`datetime.datetime`).

    Warnings
    --------
    The name of this attribute might be misleading at first. `calingen` will be
    developed in small steps, of which the first one will be to track annual
    aniversaries with the `Event` class. Nevertheless, it is planned to enable
    `calingen` to track generic recurring events. In that implementation step,
    `Event` instances will have a start and end ``datetime``.

    To ease the refactoring, this attribute is already named **start**.

    Notes
    -----
    The attribute is implemented as :class:`~django.db.models.DateTimeField`.
    """

    title = models.CharField(
        max_length=50,
        help_text=_(
            "This will be included in the generated output and is capped at 50 characters."
        ),
        verbose_name=_("Event Title"),
    )
    """The actual title of this `Event` (:py:obj:`str`).

    Notes
    -----
    The attribute is implemented as :class:`~django.db.models.CharField`.
    """

    category = models.CharField(
        max_length=18,
        choices=EventCategory.choices,
        default=EventCategory.ANNUAL_ANNIVERSARY,
        help_text=_("The category of this event."),
        verbose_name=_("Event Category"),
    )
    """The category of this `Event`.

    Notes
    -----
    Some functionalities of the `Event` class depend on the actual ``EventCategory``
    stored in this attribute.

    The attribute is implemented as :class:`~django.db.models.CharField` with
    its possible values limited by :class:`calingen.constants.EventCategory`.
    """

    class Meta:  # noqa: D106
        app_label = "calingen"
        unique_together = ["profile", "title", "start"]
        verbose_name = _("Event")
        verbose_name_plural = _("Events")

    def __str__(self):  # noqa: D105
        return "[{}] [{}] {} - {}".format(
            self.category, self.profile, self.title, self.start
        )  # pragma: nocover

    def get_absolute_url(self):
        """Return the absolute URL for instances of this model.

        Returns
        -------
        str
            The absolute URL for instances of this model.
        """
        return reverse("calingen:event-detail", args=[self.id])  # pragma: nocover

    def resolve(self, year=None):
        """Resolve this object's ``start`` for a given ``year``.

        Parameters
        ----------
        year : int, optional
            The year to resolve for. Will use the current year if not specified.

        Returns
        -------
        :class:`~calingen.interfaces.data_exchange.CalendarEntryList`
            The method returns a ``CalendarEntryList`` with entries for the
            given ``year``.
        """
        if year is None:
            year = datetime.datetime.now().year

        result = CalendarEntryList()

        # The current state of the app just includes the categories
        # ANNUAL_ANNIVERSARY and HOLIDAY. Both of them are implicitly expected
        # to have a yearly recurrence.
        # The following statement works on that assumption and simply uses the
        # specified year parameter with the (stored) values of month and day.
        result.add(
            CalendarEntry(
                self.title,
                self.category,
                datetime.date(year, self.start.month, self.start.day),
                (SOURCE_INTERNAL, self.get_absolute_url),
            )
        )

        return result


class EventForm(forms.ModelForm):
    """Used to validate input for creating and updating `Event` instances."""

    start = SplitDateTimeOptionalField(help_text=Event.start.field.help_text)

    class Meta:  # noqa: D106
        model = Event
        fields = ["category", "title", "start"]



# SPDX-License-Identifier: MIT

"""Implementation of :class:`~calingen.interfaces.plugin_api.CompilerProvider` that presents the source code of the rendered layout ready for copy/paste.

This compiler **is able to handle all types of layouts**, meaning it supports
all layout source languages (as no real compilation is performed).

This makes this compiler a suitable choice for the ``"default"`` compiler in
**django-calingen**'s settings (see :attr:`~calingen.settings.CALINGEN_COMPILER`).

Notes
-----
The compiler uses the template ``calingen/contrib/compiler_copypaste.html`` to
present the rendered source code.

This template may be substituted in a real project.
"""



# SPDX-License-Identifier: MIT

"""App-specific implementations of :class:`django.forms.widgets.Widget`."""

# Django imports
from django.forms.fields import CallableChoiceIterator
from django.forms.widgets import CheckboxSelectMultiple, HiddenInput, MultiWidget


class PluginWidget(MultiWidget):
    """Widget to render the app-specific :class:`~calingen.forms.fields.plugin.PluginField`."""

    def __init__(self, *args, **kwargs):
        widgets = (CheckboxSelectMultiple(), HiddenInput())
        super().__init__(widgets=widgets, *args, **kwargs)

    def update_available_plugins(self, choices=()):
        """Apply the choices to the multiple choice checkbox field.

        Notes
        -----
        In a normal, 1:1 relation of fields and widgets, this happens
        automatically during widget initialisation. However, using
        :class:`~django.forms.fields.MultiValueField` and
        :class:`~django.forms.widgets.MultiWidget`, this step is not performed
        for the subwidgets.

        The constructor of :class:`~calingen.forms.fields.PluginField` will take
        care of this by calling this method.
        """
        # See: django.forms.fields.ChoiceField._set_choices
        if callable(choices):
            widget_choices = CallableChoiceIterator(choices)
        else:
            widget_choices = list(choices)
        self.widgets[0].choices = widget_choices

    def decompress(self, value):
        """Expand the received value to render its components in different widgets.

        Notes
        -----
        See the corresponding :meth:`calingen.forms.fields.PluginField.compress`
        method.
        """
        if value:
            active = value.get("active", [])
            unavailable = value.get("unavailable", [])
            return [active, unavailable]

        return [None, None]



# SPDX-License-Identifier: MIT

"""The application's interfaces that are used to connect the different components.

Notes
-----
This package's code is not really specific to the Django framework. It is an
abstraction layer.

Primary focus is the provision of a plugin API, that allows the app to be
extendable with third-party applications.
"""



# SPDX-License-Identifier: MIT

"""Views related to the :class:`calingen.models.profile.Profile` model."""

# Django imports
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db import IntegrityError
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.views import generic

# app imports
from calingen.models.profile import Profile, ProfileForm
from calingen.views.mixins import ProfileIDMixin, RestrictToUserMixin


class ProfileDetailView(
    LoginRequiredMixin,
    RestrictToUserMixin,
    ProfileIDMixin,
    generic.DetailView,
):
    """Provide details of a :class:`calingen.models.profile.Profile` instance.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.DetailView`.
    """

    model = Profile
    """Required attribute to tie this view to the model."""

    context_object_name = "profile"
    """Provide a semantic name for the built-in context."""

    pk_url_kwarg = "profile_id"
    """The name of the keyword argument as provided in the app's url configuration.

    By default, this is simply ``"pk"``, but for clarity, the app's url
    configuration (:mod:`calingen.urls`) uses the more explicit ``"event_id"``.
    """

    def get_context_data(self, **kwargs):
        """Just for linting."""
        context = super().get_context_data(**kwargs)

        # add number of events derived from activated EventProvider instances
        context["provider_events_count"] = len(context["profile"].resolve())

        return context


class ProfileCreateView(LoginRequiredMixin, generic.CreateView):
    """Generic class-based view implementation to add :class:`calingen.models.profile.Profile` objects.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.CreateView`.
    """

    model = Profile
    """Required attribute to tie this view to the model."""

    fields = []
    """The fields to include into the form.

    This list is left empty, as the creation of the
    :class:`~calingen.models.profile.Profile` instance is a one-off operation,
    simply tying the instance to an existing ``User`` of the Django project.
    """

    template_name_suffix = "_create"
    """Make the view use the template ``calingen/profile_create.html``."""

    def form_valid(self, form):
        """Handle database integrity errors.

        Notes
        -----
        The view relies on an automatically generated instance of Django's
        :class:`~django.forms.ModelForm` **without** any fields. Thus,
        validation of the ``form`` is a given, as the ``request.user`` is
        injected **after** form validation.

        However, the :class:`~calingen.models.profile.Profile` has a 1:1 relation
        with the project's :setting:`AUTH_USER_MODEL`, so trying to create
        multiple instances of :class:`~calingen.models.profile.Profile` for the
        same `user` will raise an :class:`~django.db.IntegrityError`.

        If the :class:`~django.db.IntegrityError` is catched, the user will be
        redirected to his profile page.
        """
        # inject the current user into the instance
        form.instance.owner = self.request.user

        try:
            return super().form_valid(form)
        except IntegrityError:
            # The requesting user already has a profile, simply redirect him
            return redirect(
                "calingen:profile", Profile.objects.get(owner=self.request.user).id
            )


class ProfileDeleteView(
    LoginRequiredMixin,
    RestrictToUserMixin,
    ProfileIDMixin,
    generic.DeleteView,
):
    """Generic class-based view implementation to delete :class:`calingen.models.profile.Profile` objects.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.DeleteView`.
    """

    model = Profile
    """Required attribute to tie this view to the model."""

    context_object_name = "profile_item"
    """Provide a semantic name for the built-in context."""

    pk_url_kwarg = "profile_id"
    """The name of the keyword argument as provided in the app's url configuration.

    By default, this is simply ``"pk"``, but for clarity, the app's url
    configuration (:mod:`calingen.urls`) uses the more explicit ``"profile_id"``.
    """

    success_url = reverse_lazy("calingen:profile-add")
    """The URL to redirect to after successfully deleting the instance.

    Warnings
    --------
    As of now, this redirect to the page to create a profile. This **must** be
    adjusted, once the `real` estimated flows are implemented with all required
    views.
    """


class ProfileUpdateView(
    RestrictToUserMixin,
    LoginRequiredMixin,
    ProfileIDMixin,
    generic.UpdateView,
):
    """Generic class-based view implementation to add :class:`calingen.models.profile.Profile` objects.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.CreateView`.
    """

    model = Profile
    """Required attribute to tie this view to the model."""

    form_class = ProfileForm
    """Specify which form to use."""

    pk_url_kwarg = "profile_id"
    """The name of the keyword argument as provided in the app's url configuration.

    By default, this is simply ``"pk"``, but for clarity, the app's url
    configuration (:mod:`calingen.urls`) uses the more explicit ``"profile_id"``.
    """

    template_name_suffix = "_update"
    """Make the view use the template ``calingen/profile_update.html``."""

    def get_context_data(self, **kwargs):
        """Provide additional context for this view, depending on app-specific settings.

        While accessing the users :class:`~calingen.models.profile.Profile`,
        its list of ``event_provider`` is updated. If there are providers, that
        are moved from ``active`` to ``unavailable``, the user is informed using
        Django's ``messages`` framework.

        This is dependent on the setting
        :attr:`~calingen.settings.CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATION`.
        """
        if settings.CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATION == "messages":
            # get the context
            context = super().get_context_data(**kwargs)

            # evaluate, if there are newly deactivated plugins and add a message
            for deactivated_plugin in context["profile"].event_provider[
                "newly_unavailable"
            ]:
                messages.warning(
                    self.request,
                    "The following plugin is no longer available: {}".format(
                        deactivated_plugin
                    ),
                    fail_silently=True,
                )

            return context

        return super().get_context_data(**kwargs)



# SPDX-License-Identifier: MIT

"""Implementation of :class:`~calingen.interfaces.plugin_api.CompilerProvider` that directly serves HTML-based layouts and provides the source code of other rendered layouts as file download.

This compiler **is able to handle all types of layouts**, meaning it supports
all layout source languages (as no real compilation is performed).

This makes this compiler a suitable choice for the ``"default"`` compiler in
**django-calingen**'s settings (see :attr:`~calingen.settings.CALINGEN_COMPILER`).

Warnings
--------
The file extension of the (downloadable) file is determined by using the
layout's ``layout_type`` attribute (see
:class:`calingen.interfaces.plugin_api.LayoutProvider`) in combination with
:attr:`calingen.contrib.compilers.html_or_download.compiler.SOURCE_TYPE_LOOKUP`.

Notes
-----
The layout's ``layout_type`` attribute is also used to determine, if the given
source is of type ``HTML`` and can thus be served directly.
"""



# SPDX-License-Identifier: MIT

"""Implementations of :class:`django.forms.Form` in the context of rendering and compilation."""

# Python imports
from datetime import date

# Django imports
from django.forms.fields import ChoiceField, IntegerField
from django.forms.widgets import RadioSelect
from django.utils.translation import gettext_lazy as _

# app imports
from calingen.forms.generic import RequestEnabledForm
from calingen.interfaces.plugin_api import LayoutProvider


class LayoutConfigurationForm(RequestEnabledForm):
    """Layout-specific configuration form.

    Implementations of :class:`calingen.interfaces.plugin_api.LayoutProvider`
    may provide a subclass of this form to fetch layout-specific configuration
    values during the rendering process of a layout.

    Warnings
    --------
    To make this work, the Django project must provide support for
    :djangodoc:`Django's Sessions <topics/http/sessions/>`. This is in fact
    verified by :func:`calingen.checks.check_session_enabled`.

    Notes
    -----
    For an example usage, see :class:`calingen.contrib.layouts.lineatur.lineatur.LineaturForm`.
    """

    def save_configuration(self):
        """Save the layout-specific configuration in the user's ``Session``."""
        self.request.session[
            "layout_configuration"
        ] = self.cleaned_data  # pragma: nocover


class LayoutSelectionForm(RequestEnabledForm):
    """Select one of the available :class:`~calingen.interfaces.plugin_api.LayoutProvider` instances.

    Warnings
    --------
    To make this work, the Django project must provide support for
    :djangodoc:`Django's Sessions <topics/http/sessions/>`. This is in fact
    verified by :func:`calingen.checks.check_session_enabled`.

    Notes
    -----
    As all layouts require a ``target_year``, this is included here aswell.
    """

    target_year = IntegerField(
        label=_("Year"),
        help_text=_("The year to generate the layout for"),
        initial=date.today().year + 1,
        min_value=date.min.year,
        max_value=date.max.year,
    )
    """Specify the year to create the layout for."""

    layout = ChoiceField(
        label=_("Layout"),
        help_text=_("Select the layout to be used"),
        choices=LayoutProvider.list_available_plugins,
        widget=RadioSelect,
    )
    """Presentation of the available :class:`~calingen.interfaces.plugin_api.LayoutProvider`.

    Notes
    -----
    The available ``LayoutProvider`` are determined dynamically by calling
    :meth:`LayoutProvider.list_available_plugins() <calingen.interfaces.plugin_api.LayoutProvider.list_available_plugins>`
    and presented using a :class:`~django.forms.widgets.RadioSelect` widget.
    """

    def save_selection(self):
        """Save the layout selection in the user's ``Session``."""
        self.request.session["target_year"] = self.cleaned_data[
            "target_year"
        ]  # pragma: nocover
        self.request.session["selected_layout"] = self.cleaned_data[
            "layout"
        ]  # pragma: nocover



# SPDX-License-Identifier: MIT

"""Provides the web views.

These views are not model-specific (see
:mod:`calingen.views.event` and :mod:`calingen.views.profile`) nor directly
related to render and compile layouts (see :mod:`calingen.views.generation`),
but provide representations of the app's data.

Please note: The focus of calingen is to create analogous, paper-based calendar
pages. These views are provided as convenience!
"""

# Django imports
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic.base import TemplateView

# app imports
from calingen.views.mixins import AllCalendarEntriesMixin, RestrictToUserMixin


class CalendarEntryListView(
    LoginRequiredMixin,
    RestrictToUserMixin,
    AllCalendarEntriesMixin,
    TemplateView,
):
    """Provide a list view of all events in a given year.

    Notes
    -----
    The :class:`~calingen.views.mixins.CalingenUserProfileIDMixin` is not really
    required, because the view fetches the (full)
    :class:`~calingen.models.profile.Profile` instance anyway. Room for
    optimization.
    """

    template_name = "calingen/calendar_entry_list_year.html"



# SPDX-License-Identifier: MIT

"""Provides data exchange formats."""

# Python imports
import datetime
from functools import total_ordering

# external imports
from dateutil import parser

# app imports
from calingen.constants import EventCategory
from calingen.exceptions import CallingenInterfaceException

SOURCE_INTERNAL = "INTERNAL"
"""Constant for app-internal :class:`~calingen.models.event.Event` instances."""

SOURCE_EXTERNAL = "EXTERNAL"
"""Constant for entries provided by implementations of :class:`calingen.interfaces.plugin_api.EventProvider`"""


@total_ordering
class CalendarEntry:
    """Data structure to pass calendar entries around.

    Parameters
    ----------
    title : str
    category : calingen.constants.EventCategory.value
    timestamp : datetime.datetime, datetime.date, str
    source : tuple

    Attributes
    ----------
    title : str
        The actual title of the entry.
    category : str
        While this is actually a simple :py:obj:`str`, it is expected to be of
        the specified type. A lookup against
        :attr:`EventCategory.values <calingen.constants.EventCategory.values>`
        is performed, but this is not an enforced validation!

        If the ``constructor`` was called with an instance of
        :attr:`EventCategory.value <calingen.constants.EventCategory.value>`,
        this will be a `translateable` string.
    timestamp : datetime.datetime
        The paremeter accepts the specified types and will convert them to
        :py:obj:`datetime.datetime` internally.
        If a :py:obj:`str` is given, :meth:`dateutil.parser.parse` is used and
        may raise an exception, if the provided string could not be parsed.
    source : tuple
        Expected is a tuple of the form ``("INTERNAL", [int])``, where ``[int]``
        is interpreted as an ``id`` of an :class:`~calingen.models.event.Event`
        instance, or ``("EXTERNAL", [str])``, where ``[str]`` is used as a plain
        string.

        For implementations of
        :class:`~calingen.interfaces.plugin_api.EventProvider` it is recommended
        to provide its ``title`` attribute.

    Raises
    ------
    CalendarEntry.CalendarEntryException
        Raised if ``source`` was not provided as :py:obj:`tuple`
    dateutil.parser._parser.ParserError
        Raised if ``timestamp`` is provided as :py:obj:`str` and could not be
        parsed by :meth:`dateutil.parser.parse`

    Warnings
    --------
    The instance's ``source`` attribute is not included in checks for
    equality (``__eq__()``) or during hash-processing (``__hash__()``)!

    This means, that if an event is specified by an external provider and
    by using the app's :class:`calingen.models.event.Event` model, using
    the same ``title``, ``category`` and ``timestamp``, they are considered
    **equal**.

    The implementation of
    :class:`calingen.interfaces.data_exchange.CalendarEntryList` relies
    internally on a ``set``, which means the entries are unique. So, only
    one of the events will be present in the resulting ``CalendarEntryList``.

    Notes
    -----
    Instances of this class are single, self-contained entries in a calendar.
    They provide an abstraction and common interface to events provided by the
    app's user (instances of :class:`~calingen.models.event.Event`) and events
    provided by plugins (implementations of
    :class:`~calingen.interfaces.plugin_api.EventProvider`).

    However, there is no real case of using this class without and accompanying
    :class:`~calingen.interfaces.data_exchange.CalendarEntryList`. Actually,
    all ``resolve()`` operations are required to return an instance of
    :class:`~calingen.interfaces.data_exchange.CalendarEntryList` with instances
    of this class as its payload.

    As you can see, the documentation of the class's `magic methods` is kept at
    a minimum. See the source code for further details!
    """

    class CalendarEntryException(CallingenInterfaceException):
        """Class-specific exception, raised on failures in this class's methods."""

    def __init__(self, title, category, timestamp, source):
        # documentation of the costructor is in the class's docstring!
        self.title = title

        # Use the predefined category (if available)
        if category in EventCategory.values:
            self.category = EventCategory[category]
        else:
            self.category = category

        # Ensure that "timestamp" is a datetime.datetime object
        if isinstance(timestamp, datetime.datetime):
            self.timestamp = timestamp
        elif isinstance(timestamp, datetime.date):
            self.timestamp = datetime.datetime.combine(timestamp, datetime.time.min)
        else:
            # Delegate parsing to dateutil; probably breaks if unparsable
            # Django uses datetime.datetime.strptime(), dateutil.parser should
            # be "more forgiving". And while we rely on that package anyway...
            self.timestamp = parser.parse(timestamp)

        # "source" is expected to be a tuple of the the form
        # (SOURCE_INTERNAL, Event.id) or (SOURCE_EXTERNAL, EventProvider.title)
        if isinstance(source, tuple):
            self.source = source
        else:
            raise self.CalendarEntryException("source must be provided as tuple")

    def __eq__(self, other):
        """Check equality with ``other`` object."""
        # see https://stackoverflow.com/a/2909119
        # see https://stackoverflow.com/a/8796908
        if isinstance(other, CalendarEntry):
            return self.__key() == other.__key()  # pragma: nocover
        return NotImplemented

    def __lt__(self, other):
        """Provide `less than` comparison with ``other`` object."""
        # see https://stackoverflow.com/a/8796908
        if isinstance(other, CalendarEntry):
            return self.__key() < other.__key()
        return NotImplemented

    def __repr__(self):
        """Provide an instance's `representation`."""
        # see https://stackoverflow.com/a/12448200
        return "<CalendarEntry(title={}, category={}, start={}, source={})>".format(
            self.title.__repr__(),
            self.category.__repr__(),
            self.timestamp.__repr__(),
            self.source.__repr__(),
        )  # pragma: nocover

    def __str__(self):
        """Provide a string representation of the instance."""
        # see https://stackoverflow.com/a/12448200
        return "[{}] {} ({})".format(
            self.timestamp, self.title, self.category
        )  # pragma: nocover

    def __hash__(self):
        """Provide a unique representation of the instance."""
        # see https://stackoverflow.com/a/2909119
        return hash(self.__key())  # pragma: nocover

    def __key(self):
        """Provide internal representation of the instance."""
        # see https://stackoverflow.com/a/2909119
        return (self.timestamp, self.category, self.title)  # pragma: nocover


class CalendarEntryList:
    """A list of calendar entries.

    Attributes
    ----------
    _entries : set
        This attribute stores the list of calendar entries. It is implemented
        as a :py:obj:`set`, which makes it mutable and unordered. Primarily a
        :py:obj:`set` is used to ensure that all included items are **unique**.

    Warnings
    --------
    While objects of this class are intended to store / handle instances of
    :class:`~calingen.interfaces.data_exchange.CalendarEntry`, this is **not**
    enforced or validated (EAFP).

    Notes
    -----
    This class is a close companion of
    :class:`~calingen.interfaces.data_exchange.CalendarEntry` instances.

    Internally, the app expects the result of any ``resolve()`` operation to be
    an instance of this class with a set (or list) of
    :class:`~calingen.interfaces.data_exchange.CalendarEntry` instances.
    """

    class CalendarEntryListException(CallingenInterfaceException):
        """Class-specific exception, raised on failures in this class's methods."""

    def __init__(self):  # noqa: D107
        self._entries = set()

    def __len__(self):  # noqa: D105
        return len(self._entries)  # pragma: nocover

    def add(self, entry):
        """Add a :class:`~calingen.interfaces.data_exchange.CalendarEntry` to the list.

        Parameters
        ----------
        entry : CalendarEntry
            The entry to be added to this class's list.

        Raises
        ------
        CalendarEntryList.CalendarEntryListException
            Raised if no entry is provided.

        Warnings
        --------
        There is no validation of the input type!
        """
        if entry is None:
            raise self.CalendarEntryListException("An entry is required")

        self._entries.add(entry)

    def merge(self, entry_list_instance):
        """Merge two instances of ``CalendarEntryList``.

        Parameters
        ----------
        entry_list_instance : CalendarEntryList
            The instance to be merged into this one.

        Warnings
        --------
        There is no validation of the input type!

        Notes
        -----
        Providing a ``merge()`` method instead of enabling addition of objects
        of this type is a design choice and may be subject to change.
        """
        self._entries.update(entry_list_instance._entries)

    def sorted(self):
        """Return the object's ``_entries`` sorted by ``start``.

        Returns
        -------
        list
            The sorted list of
            :class:`~calingen.interfaces.data_exchange.CalendarEntry`.

        Warnings
        --------
        While this class uses internally a ``set`` to manage the entries, this
        method returns a list.

        If consuming code requires uniqueness of items, use this method as late
        as possible.
        """
        return sorted(self._entries)  # pragma: nocover



# SPDX-License-Identifier: MIT

"""Integrates the app's models into Django's admin interface."""

# Django imports
from django.contrib import admin

# app imports
from calingen.models.event import Event
from calingen.models.profile import Profile


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):  # noqa: D101
    pass


@admin.register(Profile)
class ProfileAdmin(admin.ModelAdmin):  # noqa: D101
    pass



# SPDX-License-Identifier: MIT

"""App-specific settings with their default values.

Notes
-----
These *(app-specific)* settings are *injected* into the project's ``settings``
module during startup, see this StackOverflow answer [1]_ for the general idea
and :meth:`calingen.apps.CalingenConfig.ready` for implementation details.

The actual settings are defined here with their respective default value. This
module serves as the single source of truth about app-specific settings, meaning
it defines the settings' names, as they will be used in other components of the
app.

During runtime, all settings are **read** from ``django.conf.settings`` - which
is effectively the project's ``settings`` module.

:meth:`calingen.apps.CalingenConfig.ready` checks the presence of all
app-specific settings in the project's ``settings`` module and will inject the
default values as defined in this module, if the setting is missing.

If the setting is already present in the project's ``settings`` module, it will
not be modified.

References
----------
.. [1] https://stackoverflow.com/a/47154840
"""

CALINGEN_COMPILER = {
    "default": "calingen.contrib.compilers.copy_paste.compiler.CopyPasteCompiler",
}
"""Configuration of the compiler plugins.

This setting determines the mapping of available instances of
:class:`~calingen.interfaces.plugin_api.CompilerProvider` and their association
with source files, as determined by ``layout_type`` provided by implementations
of :class:`calingen.interfaces.plugin_api.LayoutProvider`.

**Default value:** ``{ "default": "calingen.contrib.compilers.copy_paste.compiler.CopyPasteCompiler" }``

Notes
-----
More mappings may be added by specifying this setting in a project's settings
module.

The accepted *keys* are of type :py:obj:`str` and should match whatever the
project's layouts (that is: implementations of
:class:`~calingen.interfaces.plugin_api.LayoutProvider`) provide with their
respective ``layout_type`` attributes. If a given ``layout_type`` is not found
in this setting, the ``default`` compiler is used.

The expected *values* are of type :py:obj:`str` and should specify a dotted
Python path to an implementation of
:class:`calingen.interfaces.plugin_api.CompilerProvider`.

This setting is evaluated in :class:`calingen.views.generation.CompilerView`.

:func:`calingen.checks.check_config_value_compiler` verifies the presence of a
``"default"`` compiler **and** that the specified *default compiler* is
importable.
"""

CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATION = None
"""Determines, if the user should be notified about missing event providers.

**Default value:** ``None``

**Accepted values**: ``None``. ``"messages"``

Notes
-----
The user may activate implementations of
:class:`~calingen.interfaces.plugin_api.EventProvider` in his
:class:`~calingen.models.profile.Profile`. These events will then get included
while generating his calendar inlays.

However, the administrator may choose to disable certain plugins _after_ the
user has activated them. The corresponding ``EventProvider`` will then not be
available.

When the user visits his :class:`~calingen.models.profile.Profile`, he can only
select those ``EventProvider``, which are currently activated by the
administrator. Formerly activated plugins are hidden - **but still stored in
the profile instance**.

This setting controls, if the user is notified about the deactivation of a
formerly selected ``EventProvider`` using Django's ``messages`` framework.

See :meth:`calingen.views.profile.ProfileUpdateView.get_context_data` for
implementation details.

See :func:`calingen.checks.check_config_value_event_provider_notification` for
the corresponding contribution to Django's check framework.
"""



# SPDX-License-Identifier: MIT

"""App-specific contributions to :djangodoc:`Django's System check framework <topics/checks/>`.

These checks validate the values of app-specific settings and may also perform
logical checks in combination with other project settings.
"""

# Django imports
from django.conf import settings
from django.core.checks import Error
from django.utils.module_loading import import_string


def check_config_value_event_provider_notification(*args, **kwargs):
    """Verify that :attr:`~calingen.settings.CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATION` provides accepted value.

    - **Check ID**: ``calingen.e001``

    Returns
    -------
    list
        A list of :djangodoc:`Messages <topics/checks/#messages>`.
    """
    errors = []
    config_value = settings.CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATION

    if config_value is not None:
        if config_value != "messages":
            errors.append(
                Error(
                    "Unaccepted config value for CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATIONS",
                    hint=(
                        "CALINGEN_MISSING_EVENT_PROVIDER_NOTIFICATIONS has to be "
                        'either None or "messages".'
                    ),
                    id="calingen.e001",
                )
            )

    return errors


def check_session_enabled(*args, **kwargs):
    """Verify that :setting:`MIDDLEWARE` contains ``"django.contrib.sessions.middleware.SessionMiddleware"``.

    - **Check ID**: ``calingen.e002``

    Returns
    -------
    list
        A list of :djangodoc:`Messages <topics/checks/#messages>`.

    Notes
    -----
    The calingen app relies on the availability of
    :djangodoc:`Sessions <topics/http/sessions/>` while processing layouts.

    This check does only verify that Django's default Session middleware is
    active for the project. If a project substituted that middleware by a custom
    implementation, this check will fail and may be deactivated.
    """
    errors = []
    if (
        "django.contrib.sessions.middleware.SessionMiddleware"
        not in settings.MIDDLEWARE
    ):
        errors.append(
            Error(
                '"SessionMiddleware" is not in "settings.MIDDLEWARE"',
                hint=(
                    "Calingen relies on Django's Sessions. It seems as the "
                    "project does not include the SessionMiddleware. "
                    "If your project relies on a custom implementation of the "
                    "SessionMiddleware, you may safely ignore this check."
                ),
                id="calingen.e002",
            )
        )

    return errors


def check_config_value_compiler(*args, **kwargs):
    """Validate :attr:`~calingen.settings.CALINGEN_COMPILER` setting.

    - **Check ID**: ``calingen.e003`` - default compiler missing
    - **Check ID**: ``calingen.e004`` - default compiler not importable

    Returns
    -------
    list
        A list of :djangodoc:`Messages <topics/checks/#messages>`.

    Warnings
    --------
    This check does not verify the importability *(is this a word?)* of all
    configured compilers. It just ensures the presence of the ``"default"``
    compiler and its importability.

    Notes
    -----
    :attr:`~calingen.settings.CALINGEN_COMPILER` is a required configuration
    value and **must** be provided in the project's settings module. It is
    injected with a sane default value, if not provided.

    This check ensures, that ``CALINGEN_COMPILER["default"]`` is provided and
    that the provided compiler (instance of
    :class:`calingen.interfaces.plugin_api.CompilerProvider`) is importable.
    """
    errors = []
    config_value = settings.CALINGEN_COMPILER

    try:
        temp = config_value["default"]
    except (KeyError, TypeError):
        errors.append(
            Error(
                '"CALINGEN_COMPILER" does not provide a "default" compiler',
                hint=(
                    "CALINGEN_COMPILER must provide a default compiler. It is "
                    "recommended to use a compiler that can handle all types "
                    "of layouts."
                ),
                id="calingen.e003",
            )
        )
        return errors

    try:
        import_string(temp)
    except ImportError:
        errors.append(
            Error(
                'CALINGEN_COMPILER["default"] could not be imported',
                hint=(
                    "The specified compiler could not be imported. Make sure "
                    "to provide a full dotted Python path to the compiler "
                    "implementation."
                ),
                id="calingen.e004",
            )
        )

    return errors



# SPDX-License-Identifier: MIT

"""App-specific implementations of :class:`django.forms.Form` and related stuff like fields and widgets.

Warnings
--------
Siblings of :class:`django.forms.ModelForm` are (by definition) semantically
related to the respective implementation of :class:`django.db.models.Model`.

Thus, the specific forms for the app's models are included within the
:py:mod:`calingen.models` package.
"""



# SPDX-License-Identifier: MIT

"""App-specific constants, that are used throughout the application."""

# Django imports
from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _


class EventCategory(TextChoices):
    """The single source of truth for different types of events.

    Throughout the app, there are different sources for events, most notably the
    app's :class:`~calingen.models.event.Event` instances and the events as
    provided by the :class:`~calingen.interfaces.plugin_api.EventProvider`
    implementations.

    All these sources **must** provide compatible event categories.

    Notes
    -----
    Provides the accepted choices for :attr:`calingen.models.event.Event.category`.
    """

    ANNUAL_ANNIVERSARY = "ANNUAL_ANNIVERSARY", _("Annual Anniversary")
    HOLIDAY = "HOLIDAY", _("Holiday")



# SPDX-License-Identifier: MIT

"""Provides the API for plugins."""

# Python imports
from datetime import datetime

# Django imports
from django.template.loader import render_to_string
from django.utils.functional import classproperty

# app imports
from calingen.interfaces.data_exchange import (
    SOURCE_EXTERNAL,
    CalendarEntry,
    CalendarEntryList,
)


def fully_qualified_classname(class_or_instance):
    """Get the fully qualified Python path of a class (or instance).

    Parameters
    ----------
    class_or_instance : any
        The class (or instance of a class) to work on

    Returns
    -------
    str
        The fully qualified, dotted Python path of that class or instance.
    """
    if isinstance(class_or_instance, type):
        # operating on a class (NOT an instance!)
        return ".".join([class_or_instance.__module__, class_or_instance.__qualname__])

    # operating on an instance
    return ".".join(
        [
            class_or_instance.__class__.__module__,
            class_or_instance.__class__.__qualname__,
        ]
    )


class PluginMount(type):
    """Core of the plugin api.

    Notes
    -----
    Searching Google with any Django-related term usually results in at least
    five StackOverflow threads with similar issues. This failed for "django
    plugin system" (ok, there are SO threads, without providing any real
    solution).

    However, there is a blog post by Marty Alchin, dating back to 2008 [1]_,
    describing some sort of `weird Python sorcery`.
    Digging deeper, while trying to understand the purpose and possible
    application of ``metaclass`` in this context, this SO thread [2]_ was found
    (easily one of the best SO threads on a general programming technique).

    Obviously, [1]_ was ported to Python3 syntax (getting rid of ``__metaclass__``
    and apply it in the class definition of the actual mount point).

    References
    ----------
    .. [1] http://martyalchin.com/2008/jan/10/simple-plugin-framework/
    .. [2] https://stackoverflow.com/questions/100003/what-are-metaclasses-in-python
    """

    def __init__(cls, name, bases, attrs):
        """Initialize the plugin list."""
        if not hasattr(cls, "plugins"):
            # This branch only executes when processing the mount point itself.
            # So, since this is a new plugin type, not an implementation, this
            # class shouldn't be registered as a plugin. Instead, it sets up a
            # list where plugins can be registered later.
            cls.plugins = []
        else:
            # This must be a plugin implementation, which should be registered.
            # Simply appending it to the list is all that's needed to keep
            # track of it later.
            cls.plugins.append(cls)


class EventProvider(metaclass=PluginMount):
    """Mount point for plugins that provide events.

    Plugins implementing this reference **must** provide the following attribute:

    - **title** (:py:obj:`str`): The title of the provider, provided as a class
      attribute.

    Plugins implementing this reference **can** provide the following attributes
    and/or methods:

    - **entries** (:py:obj:`list`): A list of
      :class:`calingen.interfaces.data_exchange.CalendarEntry` instances.

      The default implementation of
      :meth:`~calingen.interfaces.plugin_api.EventProvider.resolve` will work
      on this :py:obj:`list`. Alternatively
      :meth:`~calingen.interfaces.plugin_api.EventProvider.resolve` may be
      re-implemented (see below).
    - **resolve(year)** : A classmethod that accepts a **year** (:py:obj:`int`)
      as parameter and returns an instance of
      :class:`calingen.interfaces.data_exchange.CalendarEntryList` containing
      the `Events` of the requested year with their corresponding meta
      information as specified by
      :class:`calingen.interfaces.data_exchange.CalendarEntry`.
    """

    @classmethod
    def list_available_plugins(cls):
        """Return the available plugins.

        Returns
        -------
        set
            The resulting :py:obj:`set` contains a 2-tuple for every plugin,
            including its `qualified classname` and its ``title`` attribute.
            The `qualified classname` is determined by
            :func:`~calingen.interfaces.plugin_api.fully_qualified_classname`.

        Notes
        -----
        Primary use case for this method is the usage in a Django view,
        specifically :class:`~calingen.models.profile.ProfileForm` uses this
        method to provide the choices of its field. That ``Form`` is then used
        in the app's views, e.g. :class:`calingen.views.profile.ProfileUpdateView`.
        """
        result = set()
        for plugin in cls.plugins:
            result.add((fully_qualified_classname(plugin), plugin.title))

        # the ``set`` is sorted on return (converting it to a list, but that's
        # fine with the current (only) consumer)
        return sorted(result, key=lambda plugin_tuple: plugin_tuple[1])

    @classmethod
    def resolve(cls, year):
        """Return a list of events.

        Parameters
        ----------
        year : int
            The year to retrieve the list of events for.

        Returns
        -------
        :class:`calingen.interfaces.data_exchange.CalendarEntryList`
            Wraps the actual events, provided as
            :class:`calingen.interfaces.data_exchange.CalendarEntry` into one
            single object.

        Notes
        -----
        This is the default implementation of the ``resolve()`` method. It makes
        some assumptions about the internal structure of an actual
        implementation of :class:`~calingen.interfaces.plugin_api.EventProvider`:

        - A class variable ``entries`` provides a :py:obj:`list` of
          :py:obj:`tuple` instances with the following structure:
          ``(title, category, recurrence)``:

            - ``title``: :py:obj:`str`
            - ``category``: :py:obj:`str`, but it is recommended to rely on a
              predefined value from :class:`calingen.constants.EventCategory`
            - ``recurrence``: :class:`dateutil.rrule.rrule`

        - A class variable ``title`` (:py:obj:`str`). This is already a required
          parameter of the :class:`~calingen.interfaces.plugin_api.EventProvider`
          implementation itsself and is used to populate the ``source``
          attribute of the returned
          :class:`~calingen.interfaces.data_exchange.CalenderEntry` instances.
        """
        result = CalendarEntryList()
        for entry in cls.entries:
            result.add(
                CalendarEntry(
                    entry[0],
                    entry[1],
                    entry[2].between(
                        datetime(year, 1, 1), datetime(year, 12, 31), inc=True
                    )[0],
                    (SOURCE_EXTERNAL, cls.title),
                )
            )
        return result


class LayoutProvider(metaclass=PluginMount):
    """Mount point for plugins that provide layouts.

    Plugins implementing this reference **must** provide the following
    attributes:

    - **title** (:py:obj:`str`): The ``title`` will be used to represent the
      plugin in the app's views. It **must be** provided as a class attribute
      resolving to a :py:obj:`str`. However, it is recommended to provide the
      following class attributes instead:

        - **name** (:py:obj:`str`): A descriptive name of the layout.
        - **paper_size** (:py:obj:`str`): The size of the paper, e.g. ``"a4"``,
          ``"a5"``, etc.
        - **orientation** (:py:obj:`str`): The orientation, e.g. ``"landscape"``
          or ``"portrait"``.

      By providing these attributes, the ``title`` can be provided automatically
      by :meth:`~calingen.interfaces.plugin_api.LayoutProvider.title` in a
      unified way throughout the application.
    - **layout_type** (:py:obj:`str`): This required class attribute specifies
      the source language of the rendered result and is used in
      :class:`calingen.views.generation.CompilerView` to determine
      which compiler is to be used to actually compile the rendered source to
      its final product.

      The actual mapping of ``layout_type`` to compiler is done in the
      app-specific setting :py:data:`~calingen.settings.CALINGEN_COMPILER`.

      ``layout_type`` may not be ``"default"``, as this is a special key used
      internally in :class:`~calingen.views.generation.CompilerView` and
      specifies the fallback compiler for the case that there is no dedicated
      compiler available for the ``layout_type`` of a layout.

    Plugins implementing this reference **can** provide implementations of the
    following methods:

    - **prepare_context(context)** (:py:obj:`dict`): This method performs
      pre-processing of the ``context``. If an actual layout needs data in a
      different structure than provided by
      :class:`calingen.views.generation.CompilerView`, this method may be
      re-implemented by the actual layout. See
      :meth:`~calingen.interfaces.plugin_api.LayoutProvider.prepare_context` for
      additional details.
    - **render(year, entries)** (:py:obj:`str`): Actually renders the layout's
      templates with the given context. This method may be re-implemented by
      actual layouts, though it is pretty generic as it is and should work for
      most use cases. It calls
      :meth:`~calingen.interfaces.plugin_api.LayoutProvider.prepare_context` for
      pre-processing of the context. It is recommended to rely on
      ``prepare_context()`` in actual layout implementations.
    """

    configuration_form = None
    """Layouts may provide a custom form to fetch configuration values.

    The specified form should be a subclass of
    :class:`calingen.forms.generation.LayoutConfigurationForm` and may implement
    custom validation / cleaning logic.
    """

    @classproperty
    def title(cls):  # noqa: B902
        """Return the plugin's human-readable title.

        Returns
        -------
        str
            The plugin's title is constructed from class attributes.

        Notes
        -----
        The ``title`` is provided as a
        :class:`~django.utils.functional.classproperty`.

        If an actual layout implementation wishes to provide its title in a
        different way, it may provide a specific implementation of this method
        or provide a plain ``title`` class attribute, resolving to a
        :py:obj:`str` (TODO: NEEDS VERIFICATION!).

        The ``classproperty`` decorator is provided by Django, so B902 is
        disabled on this.
        """
        return "{} ({}, {})".format(cls.name, cls.paper_size, cls.orientation)

    @classmethod
    def list_available_plugins(cls):
        """Return the available plugins.

        Returns
        -------
        set
            The resulting :py:obj:`set` contains a 2-tuple for every plugin,
            including its `qualified classname` and its **title** attribute.
            The `qualified classname` is determined by
            :func:`~calingen.interfaces.plugin_api.fully_qualified_classname`.

        Notes
        -----
        Primary use case for this method is the usage in a Django view,
        specifically :class:`~calingen.models.profile.ProfileForm` uses this
        method to provide the choices of its field. That ``Form`` is then used
        in the app's views, e.g. :class:`~calingen.views.profile.ProfileUpdateView`.
        """
        result = set()
        for plugin in cls.plugins:
            result.add((fully_qualified_classname(plugin), plugin.title))

        # the ``set`` is sorted on return (converting it to a list, but that's
        # fine with the current (only) consumer)
        return sorted(result, key=lambda plugin_tuple: plugin_tuple[1])

    @classmethod
    def prepare_context(cls, context):
        """Pre-process the context before rendering.

        Parameters
        ----------
        context : dict
            The context, as provided by
            :meth:`calingen.views.generation.CompilerView.get`.

        Returns
        -------
        dict
            The updated context.

        Notes
        -----
        This default implementation does nothing. ``LayoutProvider``
        implementations may override this method to apply necessary modifications
        of the ``context``.
        """
        return context  # pragma: nocover

    @classmethod
    def render(cls, context, *args, **kwargs):
        """Return the rendered source.

        Returns
        -------
        str
            The rendered source.

        Notes
        -----
        This is a very basic implementation of the (required) ``render()``
        method.
        """
        # Apply a pre-processing step to the context
        context = cls.prepare_context(context)

        return render_to_string(cls._template, context)


class CompilerProvider(metaclass=PluginMount):
    """Mount point for plugins that provide means to compile layouts to documents.

    Plugins implementing this reference **must** provide the following attributes
    and methods:

    - **title** (:py:obj:`str`): The title of the provider.
    - **get_response()** : A classmethod that accepts a :py:obj:`str`, which
      will be the rendered layout source as provided by
      :meth:`LayoutProvider.render() <calingen.interfaces.plugin_api.LayoutProvider.render>`.
      The method **must return** a valid
      :djangoapi:`Django Response object <request-response/#httpresponse-objects>`.
    """

    @classmethod
    def list_available_plugins(cls):
        """Return the available plugins.

        Returns
        -------
        set
            The resulting :py:obj:`set` contains a 2-tuple for every plugin,
            including its `qualified classname` and its **title** attribute.
            The `qualified classname` is determined by
            :func:`~calingen.interfaces.plugin_api.fully_qualified_classname`.
        """
        result = set()
        for plugin in cls.plugins:
            result.add((fully_qualified_classname(plugin), plugin.title))

        # the ``set`` is sorted on return (converting it to a list, but that's
        # fine with the current (only) consumer)
        return sorted(result, key=lambda plugin_tuple: plugin_tuple[1])

    @classmethod
    def get_response(cls, source, *args, **kwargs):
        """Get the compiler's HTTP response.

        This is the compiler's main interface to other components of the app.

        It should be used to actually trigger the compilation of the provided
        source and then return a valid
        :djangoapi:`Django Response object <request-response/#httpresponse-objects>`.

        Parameters
        ----------
        source : str
            The source as provided by
            :meth:`LayoutProvider.render() <calingen.interfaces.plugin_api.LayoutProvider.render>`.

        Returns
        -------
        :djangoapi:`Django Response object <request-response/#httpresponse-objects>`

        Notes
        -----
        This method is called from
        :meth:`CompilerView's get() method <calingen.views.generation.CompilerView.get>`
        with an additional keyword argument ``layout_type``, exposing the
        :attr:`~calingen.interfaces.plugin_api.LayoutProvider.layout_type`
        attribute of the layout.
        """
        raise NotImplementedError(
            "Has to be implemented by the actual provider"
        )  # pragma: nocover



# SPDX-License-Identifier: MIT

"""View implementations with generic functionality."""


# Django imports
from django.views.generic.edit import FormView


class RequestEnabledFormView(FormView):
    """Basically just a :class:`~django.views.generic.edit.FormView` that makes the ``request`` available on the form.

    Notes
    -----
    This view implementation is intended to be used together with
    :class:`calingen.forms.generic.RequestEnabledForm`. In fact, an actual view
    will subclass this view and use a form that is a subclass of
    :class:`~calingen.forms.generic.RequestEnabledForm`.
    """

    def get_form_kwargs(self, *args, **kwargs):
        """Add the ``request`` object to the form during initialization."""
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["request"] = self.request
        return kwargs



# SPDX-License-Identifier: MIT

"""Implementation of :class:`~calingen.interfaces.plugin_api.EventProvider` that provide German holidays, splitted by federate states.

In fact, it is not only *one* implementation of
:class:`~calingen.interfaces.plugin_api.EventProvider`, but 17.
"""


default_app_config = (
    "calingen.contrib.providers.german_holidays.apps.CalingenProviderGermanHolidays"
)
"""The path to the app's default configuration class.

Consider this *legacy code*. See
:djangoapi:`Django's documentation<applications/#configuring-applications>` for
details.
"""



# SPDX-License-Identifier: MIT

"""A pluggable Django application to generate analogous, legacy, paper-based calendar inlays."""

__author__ = "Mischback"
"""The current project owner."""

__app_name__ = "django-calingen"
"""The name of the application."""

__version__ = "1.1.1"
"""The current version."""

default_app_config = "calingen.apps.CalingenConfig"
"""The path to the app's default configuration class.

Consider this *legacy code*. See
:djangoapi:`Django's documentation<applications/#configuring-applications>` for
details.
"""



# SPDX-License-Identifier: MIT

"""App-specific implementations of :class:`django.forms.fields.Field`."""

# Django imports
from django.core.exceptions import ValidationError
from django.forms.fields import (
    CharField,
    MultipleChoiceField,
    MultiValueField,
    SplitDateTimeField,
)

# app imports
from calingen.forms.widgets import PluginWidget


class SplitDateTimeOptionalField(SplitDateTimeField):
    """App-specific implementation of Django's ``SplitDateTimeField``.

    While Django's default implementation of
    :class:`django.forms.fields.SplitDateTimeField` requires a `valid` value for
    both, ``date`` and ``time``, `calingen` will be happy with just a date and
    then assume that the ``time`` is ``"00:00"``.

    Warnings
    --------
    The field will accept an *empty value* for the ``time`` part, but will
    actually provide a default value of ``"00:00"`` on processing the form.

    This means, that the database will contain a :py:obj:`datetime.datetime`
    object with ``time``.
    """

    def clean(self, value):
        """Catch validation errors of the ``time`` part and provide a fallback.

        Parameters
        ----------
        value : any
            The value to be cleaned.
            As this is derived from a :class:`django.forms.fields.MultiValueField`,
            the ``value`` will actually be a tuple, containing the values for
            the ``date`` and the ``time`` part.
            This method just provides a fallback for the ``time`` part and then
            uses ``super().clean(value)`` to chain further validation/cleaning
            steps.
        """
        try:
            if self.fields[1].to_python(value[1]) is None:
                value[1] = "00:00"
        except ValidationError:
            value[1] = "00:00"

        return super().clean(value)


class CalingenListField(CharField):
    """Misuse a ``CharField`` to handle a list of strings."""

    def prepare_value(self, value):
        """Create a comma-seperated list of strings from a Python list.

        Parameters
        ----------
        value : list

        Returns
        -------
        str
            The ``value`` :py:obj:`list` is split and provided as a
            :py:obj:`str` of comma-seperated values.

        Notes
        -----
        In Django's normal form rendering workflow, this method is called to
        process a value for display in a widget. However, for a Django
        :class:`django.forms.fields.MultiValueField`, the method is not called
        for `sub fields`.

        As this field is specifically intended to work with
        :class:`~calingen.forms.fields.PluginField`,
        :meth:`its prepare_value() <calingen.forms.fields.PluginField.prepare_value>`
        method calls this method.
        """
        return ", ".join(value)  # pragma: nocover

    def to_python(self, value):
        """Convert a comma-seperated list of values to a Python list.

        Parameters
        ----------
        value : str

        Returns
        -------
        list
            The ``value`` :py:obj:`str` is split and converted in an actual
            Python :py:obj:`list`.

        Notes
        -----
        While :meth:`~calingen.forms.fields.CalingenListField.prepare_value` has
        to be called specifically, this very method is called automatically
        during form processing.
        """
        if not value:
            return []

        return [item.strip() for item in value.split(",")]


class PluginField(MultiValueField):
    """Custom field to manage JSON objects related to plugins.

    :class:`calingen.models.profile.Profile` is used to manage the user's
    selection of :class:`~calingen.interfaces.plugin_api.EventProvider` plugins.

    :meth:`calingen.models.profile.Profile.event_provider` handles this
    internally and maintains a specific JSON schema (without formally defining
    it), tracking ``active`` and ``unavailable`` instances of
    :class:`~calingen.interfaces.plugin_api.EventProvider`.

    While the model stores this as a :class:`~django.db.models.JSONField`, this
    custom field is used to present it in a form context.

    Warnings
    --------
    Currently this is very tightly connected to the actual implementation in
    :meth:`calingen.models.profile.Profile.event_provider` and will work only
    with that model / field *out of the box*.

    Additionally, it will only work with
    :class:`calingen.forms.widgets.PluginWidget`, because this class's
    :meth:`~calingen.forms.fields.PluginField.compress` method needs a
    counterpart in the widget's
    :meth:`~calingen.forms.widgets.PluginWidget.decompress` method.

    Notes
    -----
    The implementation extends
    :class:`django.forms.fields.MultiValueField` and provides the ``active``
    plugins in a :class:`django.forms.fields.MultipleChoiceField` and tracks the
    ``unavailable`` plugins in a
    :class:`calingen.forms.fields.CalingenListField`.
    """

    widget = PluginWidget
    """Use :class:`~calingen.forms.widgets.PluginWidget` as the fields widget."""

    def __init__(self, *args, choices=(), **kwargs):
        self.choices = choices

        fields = (MultipleChoiceField(), CalingenListField())

        super().__init__(fields=fields, *args, **kwargs)

        self.fields[0].choices = self.choices
        self.widget.update_available_plugins(self.choices)

    def prepare_value(self, value):
        """Prepare the value / values for display.

        Parameters
        ----------
        value : dict
            ``value`` is actually a JSON object as fetched from
            :meth:`Profile.event_provider <calingen.models.profile.Profile.event_provider>`.

        Notes
        -----
        This method *splits* the ``value`` to be used by the widgets.

        Additionally it ensures, that
        :meth:`CalingenListField.prepare_value <calingen.forms.fields.CalingenListField.prepare_value>`
        is called to handle the ``unavailable`` plugins.
        """
        prepared_list_field = self.fields[1].prepare_value(value["unavailable"])
        value["unavailable"] = prepared_list_field

        return super().prepare_value(value)

    def compress(self, data_list):
        """Compress the values of multiple fields into one object for the ORM layer.

        Parameters
        ----------
        data_list : list
            A :py:obj:`list` of values as provided by the multiple widgets of
            :class:`~calingen.forms.widgets.PluginWidget`.

        Returns
        -------
        obj
            Converts the ``data_list`` back to a object, satisfying
            :meth:`Profile.event_provider's <calingen.models.profile.Profile.event_provider>`
            JSON scheme.

        Notes
        -----
        See the corresponding :meth:`calingen.forms.widgets.PluginWidget.decompress`
        method.
        """
        if data_list:
            result = {}
            result["active"] = data_list[0]
            result["unavailable"] = data_list[1]
            return result
        return None



# SPDX-License-Identifier: MIT

"""Provide the application configuration for Django."""

# Django imports
from django.apps import AppConfig

APP_CONFIG_NAME = "CALINGEN_PROVIDER_GERMAN_HOLIDAYS"
"""The name of the app-specific configuration option."""


class CalingenProviderGermanHolidays(AppConfig):
    """Application-specific configuration class, as required by Django.

    This sub-class of Django's `AppConfig` provides application-specific
    information to be used in Django's application registry (see
    :djangoapi:`applications/#configuring-applications`).
    """

    name = "calingen.contrib.providers.german_holidays"
    verbose_name = "CalInGen Provider: German Holidays"

    def ready(self):
        """Apply app-specific stuff.

        Notes
        -----
        This method is executed when the application is (completely) loaded.

        It will import the actual implementations of
        :class:`~calingen.interfaces.plugin_api.EventProvider`, making them
        available in **django-calingen**.
        """
        # local imports
        from .provider import (  # noqa: F401
            BadenWuerttemberg,
            Bayern,
            Berlin,
            Brandenburg,
            Bremen,
            GermanyFederal,
            Hamburg,
            Hessen,
            MecklenburgVorpommern,
            Niedersachsen,
            NordrheinWestphalen,
            RheinlandPfalz,
            Saarland,
            Sachsen,
            SachsenAnhalt,
            SchleswigHolstein,
            Thueringen,
        )



# SPDX-License-Identifier: MIT

"""App-specific templatetags for escaping strings in templates.

Calingen is intended to support various types of layouts, meaning the app is
able to support several *target languages* of templates (e.g. TeX, HTML, ...).

Different *target languages* require different strings to be escaped. All of
these are provided with this module.

Notes
-----
The required regular expressions to escape strings are compiled at module level
and may be re-used. This should boost performance during template processing.

An actual ``filter`` for HTML is not provided, this should already be covered
by Django's codebase.
"""

# Python imports
import re

# Django imports
from django import template
from django.template.defaultfilters import stringfilter

register = template.Library()

RE_BACKSLASH = re.compile(r"\\")
"""RegEx to *find* backslashes (``\``)."""  # noqa: W605

RE_EXP = re.compile(r"\^")
"""RegEx to *find* carets (``^``)."""

RE_TILDE = re.compile(r"~")
"""RegEx to *find* tildes (``~``)."""

RE_TEX_CHARS = re.compile(r"([$#&%_{}])")
"""RegEx to *find* a bunch of TeX special characters (``$#&%_{}``)."""

RE_TEX_BACKSLASH_FIX = re.compile(r"(\\backslash)")
"""This RegEx is used to **fix** *backslash* escaping in TeX templates."""


@register.filter
@stringfilter
def escape_tex(value):
    r"""Escape TeX control characters/sequences.

    Parameters
    ----------
    value : str
        The string to be cleaned from TeX control sequences.

    Returns
    -------
    str
        The cleaned / escaped string.

    Notes
    -----
    Special characters / sequences to be processed:

    ``& % $ # _ { } ~ ^ \``

    The list of special TeX characters is taken from StackOverflow
    (see https://tex.stackexchange.com/a/34586). Actually, the same thread has
    the (Pearl) regular expressions to substitute these special characters
    aswell (see https://tex.stackexchange.com/a/119383).
    """
    # The order of substitution *does* matter!
    value = RE_BACKSLASH.sub(r"\\backslash", value)
    value = RE_TEX_CHARS.sub(r"\\\1", value)
    value = RE_EXP.sub(r"\^{}", value)
    value = RE_TILDE.sub(r"\\texttt{\~{}}", value)
    value = RE_TEX_BACKSLASH_FIX.sub(r"$\1$", value)

    return value



# SPDX-License-Identifier: MIT

"""App-specific implementation of :class:`django.db.models.QuerySet`."""


# Django imports
from django.db.models import QuerySet


class CalingenQuerySet(QuerySet):
    """Base class for all app-specific ``QuerySet`` implementations.

    Notes
    -----
    This :class:`~django.db.models.QuerySet` implementation provides
    app-specific augmentations, most notable the ability to filter the resulting
    set by the class's ``owner`` attribute, which is a reference to
    :setting:`AUTH_USER_MODEL`.

    The app's :class:`~django.db.models.QuerySet` implementations are applied
    by the model's specific implementation of :class:`django.db.models.Manager`,
    which is available as attribute ``calingen_manager`` on every model of the
    `calingen` app.
    """

    def default(self):
        """Return a queryset with annotations.

        Returns
        -------
        :class:`django.db.models.QuerySet`
            The annotated queryset.

        Warnings
        --------
        Currently, this method does nothing on its own, but is kept to keep the
        app's specific QuerySets consistent.
        """
        return self  # pragma: nocover

    def filter_by_user(self, user):
        """Return a queryset filtered by the ``owner`` attribute.

        Parameters
        ----------
        user :
            An instance of the project's user model, as specified by
            :setting:`AUTH_USER_MODEL`.

        Returns
        -------
        :class:`django.db.models.QuerySet`
            The filtered queryset.

        Notes
        -----
        Effectively, this method is used to ensure, that any user may only
        access objects, which are owned by him. This is the app's way of
        ensuring `row-level permissions`, because only owners are allowed to
        view (and modify) their events.
        """
        raise NotImplementedError(
            "Must be implemented by actual QuerySet"
        )  # pragma: nocover



# Generated by Django 4.1.5 on 2023-01-06 13:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('calingen', '0001_initial'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='event',
            unique_together={('profile', 'title', 'start')},
        ),
    ]



# SPDX-License-Identifier: MIT

"""Related but optional functionality to django-calingen.

This package includes working example implementations of
:class:`~calingen.interfaces.plugin_api.EventProvider`,
:class:`~calingen.interfaces.plugin_api.LayoutProvider` and
:class:`~calingen.interfaces.plugin_api.CompilerProvider`.

All of these implementation are fully working and may be used in a real project
setup of **django-calingen**.

They are derived from the author's own requirements and may not fully meet all
users' expectations.

However, they may be used as a reference for implementations of the app's
plugin interfaces (see :ref:`calingen-dev-doc-plugins-label` for the relevant
part of this documentation).
"""



# SPDX-License-Identifier: MIT

"""App-specific user profile.

Beside the actual :class:`calingen.models.profile.Profile` model, this module
contains the related implementations of :class:`django.db.models.QuerySet`,
:class:`django.db.models.Manager` and :class:`django.forms.ModelForm`.
"""

# Python imports
import datetime

# Django imports
from django import forms
from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

# app imports
from calingen.forms.fields import PluginField
from calingen.interfaces.data_exchange import CalendarEntryList
from calingen.interfaces.plugin_api import EventProvider
from calingen.models.queryset import CalingenQuerySet


class ProfileQuerySet(CalingenQuerySet):
    """App-specific implementation of :class:`django.db.models.QuerySet`.

    Notes
    -----
    This :class:`~django.db.models.QuerySet` implementation provides
    app-specific augmentations.

    The provided methods augment/extend the retrieved
    :class:`calingen.models.profile.Profile` instances by annotating them with
    additional information.
    """

    def default(self):
        """Return a :class:`~django.db.models.QuerySet` with annotations.

        Returns
        -------
        :class:`~django.db.models.QuerySet`
            The annotated queryset.

        Notes
        -----
        The following annotations are provided by default:

        - :meth:`~calingen.models.profile.ProfileQuerySet._owner`
        - :meth:`~calingen.models.profile.ProfileQuerySet._event_count`
        """
        return self._owner()._event_count()

    def filter_by_user(self, user):
        """Filter the result set by the objects' :attr:`owners <calingen.models.profile.Profile.owner>`.

        Parameters
        ----------
        user :
            An instance of the project's user model, as specified by
            :setting:`AUTH_USER_MODEL`.

        Returns
        -------
        :class:`django.db.models.QuerySet`
            The filtered queryset.

        Notes
        -----
        Effectively, this method is used to ensure, that any user may only
        access objects, which are owned by him. This is the app's way of
        ensuring `row-level permissions`, because only owners are allowed to
        view (and modify) their events.
        """
        return self.filter(owner=user)

    def _event_count(self):
        """Annotate each instance with the count of associated :class:`~calingen.models.event.Event` instances.

        Returns
        -------
        :class:`~django.db.models.QuerySet`
            The annotated queryset. The annotation will be available as
            ``event_count``.

        Notes
        -----
        This annotation is provided in
        :meth:`~calingen.models.profile.ProfileQuerySet.default`.
        """
        return self.annotate(event_count=models.Count("events"))

    def _owner(self):
        """Make :attr:`Profile.owner <calingen.models.profile.Profile.owner>` available.

        Returns
        -------
        :class:`~django.db.models.QuerySet`
            Instances of :class:`~calingen.models.profile.Profile` will have
            their :attr:`~calingen.models.profile.Profile.owner` available
            without another database query.

        Notes
        -----
        This method makes the associated project user (specified by
        :setting:`AUTH_USER_MODEL` and stored in
        :attr:`Profile.owner <calingen.models.profile.Profile.owner>`) available.

        This annotation is provided in
        :meth:`~calingen.models.profile.ProfileQuerySet.default`.
        """
        return self.select_related("owner")  # pragma: nocover


class ProfileManager(models.Manager):
    """App-/model-specific implementation of :class:`django.db.models.Manager`.

    Notes
    -----
    This :class:`~django.db.models.Manager` implementation is used as an
    **additional** manager of :class:`~calingen.models.profile.Profile`
    (see :attr:`calingen.models.profile.Profile.calingen_manager`).

    This implementation inherits its functionality from
    :class:`django.db.models.Manager` and provides identical funtionality.
    Furthermore, it augments the retrieved objects with additional attributes,
    using the custom :class:`~django.db.models.QuerySet` implementation
    :class:`~calingen.models.profile.ProfileQuerySet`.
    """

    def get_queryset(self):
        """Use the app-/model-specific :class:`~calingen.models.profile.ProfileQuerySet` by default.

        Returns
        -------
        :class:`django.models.db.QuerySet`
            This queryset is provided by
            :class:`calingen.models.profile.ProfileQuerySet` and applies its
            :meth:`~stockings.models.profile.ProfileQuerySet.default`
            method. The retrieved objects will be annotated with additional
            attributes.
        """
        return ProfileQuerySet(self.model, using=self._db).default()  # pragma: nocover

    def get_profile(self, user):
        """Retrieve the :class:`~calingen.models.profile.Profile` associated with a ``User`` instance.

        Parameters
        ----------
        user :
            An instance of the project's user model, as specified by
            :setting:`AUTH_USER_MODEL`.

        Returns
        -------
        :class:`calingen.models.profile.Profile`
            The ``Profile`` associated with the given ``user``.
        """
        try:
            return self.get(owner=user)
        except self.model.DoesNotExist:
            return None


class Profile(models.Model):
    """Represents the app-specific profile.

    Warnings
    --------
    The class documentation only includes code that is actually shipped by the
    `calingen` app. Inherited attributes/methods (provided by Django's
    :class:`~django.db.models.Model`) are not documented here.
    """

    objects = models.Manager()
    """The model's default manager.

    The default manager is set to :class:`django.db.models.Manager`, which is
    the default value. In order to add the custom :attr:`calingen_manager` as
    an *additional* manager, the default manager has to be provided explicitly
    (see :djangodoc:`topics/db/managers/#default-managers`).
    """

    calingen_manager = ProfileManager()
    """App-/model-specific manager, that provides additional functionality.

    This manager is set to
    :class:`calingen.models.profile.ProfileManager`. Its implementation provides
    augmentations of `Profile` objects, by annotating them on database level.
    This will reduce the number of required database queries, if attributes of
    the object are accessed.

    The manager has to be used explicitly.
    """

    _event_provider = models.JSONField(
        default=dict, blank=True, verbose_name=_("Event Provider")
    )
    """Manage :class:`~calingen.interfaces.plugin_api.EventProvider` plugins for this profile.

    Notes
    -----
    This is implemented as a :class:`~django.db.models.JSONField` and should be
    interfaced by its custom
    :meth:`~calingen.models.profile.Profile.event_provider` getter and setter
    methods, provided as a ``property``.
    """

    owner = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("Owner")
    )
    """Reference to a Django `User`.

    Notes
    -----
    This is implemented as a :class:`~django.db.models.OneToOneField` with
    ``on_delete=CASCADE``, meaning: if the referenced `User` object is deleted,
    the referencing `Profile` object is discarded aswell.

    To keep this application as pluggable as possible, the referenced class is
    dependent on :setting:`AUTH_USER_MODEL`. With this implementation, the
    project may substitute the :class:`~django.contrib.auth.models.User` model
    provided by Django without breaking any functionality in `calingen` (see
    :djangodoc:`Reusable Apps and AUTH_USER_MODEL <topics/auth/customizing/#reusable-apps-and-auth-user-model>`).
    """

    class Meta:  # noqa: D106
        app_label = "calingen"
        verbose_name = _("Profile")
        verbose_name_plural = _("Profiles")

    def __str__(self):  # noqa: D105
        return "[Profile] {}".format(self.owner)  # pragma: nocover

    def get_absolute_url(self):
        """Return the absolute URL for instances of this model.

        Returns
        -------
        str
            The absolute URL for instances of this model.
        """
        return reverse("calingen:profile", args=[self.id])  # pragma: nocover

    def resolve(self, year=None):
        """Combine all event providers results for a given year into one :class:`~calingen.interfaces.data_exchange.CalendarEntryList`.

        Parameters
        ----------
        year : int, optional
            The year to use for resolving the
            :class:`~calingen.interfaces.plugin_api.EventProvider`.

        Returns
        -------
        :class:`~calingen.interfaces.data_exchange.CalendarEntryList`
            A single instance including all events from all active providers.
        """
        if year is None:
            year = datetime.datetime.now().year

        result = CalendarEntryList()
        for provider in self.event_provider["active"]:
            provider_instance = import_string(provider)
            result.merge(provider_instance.resolve(year))

        return result

    @property
    def event_provider(self):
        """Get and set the list of :class:`~calingen.interfaces.plugin_api.EventProvider`.

        This method is used to access the ``Profile``'s ``EventProvider``. The
        returned object will have three attributes, ``active``, ``unavailable``
        and ``newly_unavailable``: ``active`` contains the selected ``EventProvider``
        plugins, **that are currently active in this project**. If the user
        had selected ``EventProvider`` plugins, that are currently not available
        in this project, i.e. because they were deactivated by the administrator,
        these plugins are moved to ``unavailable``. Also, these plugins are
        included in ``newly_unavailable`` and then picked up in the
        :meth:`~calingen.views.profile.ProfileUpdateView.get_context_data`
        method to provide messages to the user.

        Notes
        -----
        While the implementation of the ``getter`` involves some logic to
        perform the operations as described above, the ``setter`` simply applies
        the provided value to
        :attr:`~calingen.models.profule.Profile._event_provider`. This means,
        that if the user did not actually **update** its profile, the operation
        performed by the ``getter`` is simply discarded.
        """
        raw = self._event_provider

        provider = [p[0] for p in EventProvider.list_available_plugins()]

        active = []
        unavailable = []
        newly_unavailable = []

        for item in raw.get("active", []):
            if item in provider:
                active.append(item)
            else:
                newly_unavailable.append(item)
                unavailable.append(item)

        for item in raw.get("unavailable", []):
            if item in provider:
                active.append(item)
            else:
                unavailable.append(item)

        result = {}
        result["active"] = active
        result["unavailable"] = unavailable
        result["newly_unavailable"] = newly_unavailable

        return result

    @event_provider.setter
    def event_provider(self, value):  # pragma: nocover
        if value is None:
            value = dict()
        self._event_provider = value
        self.save()


class ProfileForm(forms.ModelForm):
    """Used to validate input for creating and updating :class:`~calingen.models.profile.Profile` instances."""

    event_provider = PluginField(
        required=False, choices=EventProvider.list_available_plugins
    )
    """Manually add the model's ``property`` to the form.

    Notes
    -----
    :class:`~django.forms.ModelForm` only includes actual model fields.
    :attr:`calingen.models.profile.Profile.event_provider` is actually only a
    ``property`` to interface
    :attr:`~calingen.models.profile.Profile._event_provider`.

    The ``property`` is added here and then applied in the form's
    :meth:`~calingen.models.profile.ProfileForm.save` method.
    """

    def __init__(self, *args, **kwargs):  # pragma: nocover
        instance = kwargs.get("instance", None)
        if instance is not None:
            kwargs["initial"] = {"event_provider": instance.event_provider}
        super().__init__(*args, **kwargs)

    def save(self, *args, **kwargs):  # pragma: nocover
        """Inject the value of ``event_provider`` into the instance and save it."""
        self.instance.event_provider = self.cleaned_data["event_provider"]
        return super().save(*args, **kwargs)

    class Meta:  # noqa: D106
        model = Profile
        fields = ["event_provider"]



# SPDX-License-Identifier: MIT

"""Views in the context of rendering and compilation of layouts."""

# Python imports
from datetime import date
from logging import getLogger

# Django imports
from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.module_loading import import_string
from django.views.generic.base import ContextMixin, View

# app imports
from calingen.exceptions import CalingenException
from calingen.forms.generation import LayoutSelectionForm
from calingen.views.generic import RequestEnabledFormView
from calingen.views.mixins import AllCalendarEntriesMixin, RestrictToUserMixin

# get a module level logger
logger = getLogger(__name__)


class CompilerView(
    LoginRequiredMixin, RestrictToUserMixin, AllCalendarEntriesMixin, ContextMixin, View
):
    """Render the selected layout and pass the result to a compiler."""

    class NoLayoutSelectedException(CalingenException):
        """Raised if there is no selected layout in the user's ``Session``."""

    def get(self, *args, **kwargs):
        """Render the selected layout and call the compiler on the result.

        The actual response to the GET request is provided by the implementation
        of :meth:`calingen.interfaces.plugin_api.CompilerProvider.get_response`.

        Notes
        -----
        If there is no selected layout in the user's ``Session``, a redirect to
        :class:`calingen.views.generation.LayoutSelectionView` is performed.

        The method retrieves the
        :class:`compiler instance <calingen.interfaces.plugin_api.CompilerProvider>`
        from the project's settings module
        (:attr:`~calingen.settings.CALINGEN_COMPILER`). It will resort to the
        configured ``"default"`` compiler, if no specific compiler for the
        selected ``layout_type`` (as defined by the implementation of
        :class:`~calingen.interfaces.plugin_api.LayoutProvider`) is set or if
        the specified compiler can not be imported. In that case a log message
        (of level warn) is emitted.
        """
        try:
            layout = self._get_layout()
        except self.NoLayoutSelectedException:
            return redirect("calingen:layout-selection")

        render_context = self._prepare_context(*args, **kwargs)
        rendered_source = layout.render(render_context)

        try:
            compiler = import_string(settings.CALINGEN_COMPILER[layout.layout_type])
        except KeyError:
            compiler = import_string(settings.CALINGEN_COMPILER["default"])
        except ImportError:
            logger.warn(
                "Could not import {}, using default compiler".format(
                    settings.CALINGEN_COMPILER[layout.layout_type]
                )
            )
            compiler = import_string(settings.CALINGEN_COMPILER["default"])

        return compiler.get_response(rendered_source, layout_type=layout.layout_type)

    def _get_layout(self):
        """Return the :class:`~calingen.interfaces.plugin_api.LayoutProvider` implementation.

        Notes
        -----
        If there is no selected layout in the user's ``Session``, a custom
        exception will cause a redirect to the user's profile overview.
        """
        selected_layout = self.request.session.pop("selected_layout", None)
        if selected_layout is None:
            # This is most likely an edge case: The view is accessed with a
            # GET request without a selected layout stored in the user's session.
            # This could be caused by directly calling this view's url.
            # Just redirect to the layout selection.
            raise self.NoLayoutSelectedException()

        return import_string(selected_layout)

    def _prepare_context(self, *args, **kwargs):
        """Prepare the context passed to the layout's rendering method.

        Notes
        -----
        The ``context`` that is passed to the layout's ``render()`` method
        contains the following ``keys``:

        - ``target_year``: The year to create the layout for.
        - ``layout_configuration``: If the layout provides a custom
          implementation of :class:`calingen.forms.generation.LayoutConfigurationForm`,
          the fetched values will be provided here.
        - ``entries``: All calendar entries of the user's profile, resolved to
          the ``target_year``, provided as a
          :class:`calingen.interfaces.data_exchange.CalendarEntryList` object.
        """
        target_year = self.request.session.pop("target_year", date.today().year)
        layout_configuration = self.request.session.pop("layout_configuration", None)

        return self.get_context_data(
            target_year=target_year, layout_configuration=layout_configuration, **kwargs
        )


class LayoutConfigurationView(LoginRequiredMixin, RequestEnabledFormView):
    """Show the (optional) configuration form for the selected layout.

    Warnings
    --------
    This view is not restricted to users with a
    :class:`Calingen Profile <calingen.models.profile.Profile>` and can be
    accessed by any user of the project.

    However, on actual generation and compilation of the output, a
    ``Profile`` is required.

    Notes
    -----
    This is just the view to show and process the layout's implementation of
    :class:`calingen.forms.generation.LayoutConfigurationForm`.
    """

    template_name = "calingen/layout_configuration.html"
    success_url = reverse_lazy("calingen:compilation")

    class NoConfigurationFormException(CalingenException):
        """Raised if the selected layout does not have a ``configuration_form``."""

    class NoLayoutSelectedException(CalingenException):
        """Raised if there is no selected layout in the user's ``Session``."""

    def form_valid(self, form):
        """Trigger saving of the configuration values in the user's ``Session``."""
        form.save_configuration()
        return super().form_valid(form)

    def get(self, request, *args, **kwargs):
        """Handle a GET request to the view.

        While processing the request, it is determined if the selected
        implementation of :class:`calingen.interfaces.plugin_api.LayoutProvider`
        uses a custon ``configuration_form``.

        If no custom configuration is implemented by the layout, the request
        is redirected to the generator.

        Notes
        -----
        Determining the ``configuration_form`` is done implicitly while
        traversing the view's hierarchy during processing the request. Several
        methods are involved, but at some point
        :meth:`~calingen.views.generation.LayoutConfigurationView.get_form_class` is
        called, which will raise an exceptions that is handled here.

        If there is no selected layout in the user's ``Session``, a redirect to
        :class:`calingen.views.generation.LayoutSelectionView` is performed.
        """
        try:
            return super().get(request, *args, **kwargs)
        except self.NoConfigurationFormException:
            # As no layout specific configuration form is required, directly
            # redirect to the compilation.
            return redirect("calingen:compilation")
        except self.NoLayoutSelectedException:
            # This is most likely an edge case: The view is accessed with a
            # GET request without a selected layout stored in the user's session.
            # This could be caused by directly calling this view's url.
            # Just redirect to the layout selection.
            return redirect("calingen:layout-selection")

    def get_form_class(self):
        """Retrieve the layout's configuration form.

        Notes
        -----
        Implementations of :class:`calingen.interfaces.plugin_api.LayoutProvider`
        may provide a class attribute ``configuration_form`` with a subclass of
        :class:`calingen.forms.generation.LayoutConfigurationForm`.

        If ``configuration_form`` is omitted, a custom exception is raised, that
        will be handled in
        :meth:`~calingen.views.generation.LayoutConfigurationView.get`.

        If there is no selected layout in the user's ``Session``, a different
        custom exception will cause a redirect to the user's profile overview.
        """
        selected_layout = self.request.session.get("selected_layout", None)
        if selected_layout is None:
            raise self.NoLayoutSelectedException()

        layout = import_string(selected_layout)
        if layout.configuration_form is None:
            raise self.NoConfigurationFormException()

        return layout.configuration_form


class LayoutSelectionView(LoginRequiredMixin, RequestEnabledFormView):
    """Provide a list of availabe layouts.

    Warnings
    --------
    This view is not restricted to users with a
    :class:`Calingen Profile <calingen.models.profile.Profile>` and can be
    accessed by any user of the project.

    However, on actual generation and compilation of the output, a
    ``Profile`` is required.

    Notes
    -----
    This is just the view to show and process the
    :class:`calingen.forms.generation.LayoutSelectionForm`.

    Relevant logic, that affects the actual creation, rendering and compilation
    of layouts is provided in the corresponding
    :class:`~django.forms.Form` instance.
    """

    template_name = "calingen/layout_selection.html"
    form_class = LayoutSelectionForm
    success_url = reverse_lazy("calingen:layout-configuration")

    def form_valid(self, form):
        """Trigger saving of the selected value in the user's ``Session``."""
        form.save_selection()
        return super().form_valid(form)



# SPDX-License-Identifier: MIT

"""Application configuration as required by Django."""

# Python imports
import logging

# Django imports
from django.apps import AppConfig
from django.conf import settings
from django.core.checks import register as register_check

# get a module-level logger
logger = logging.getLogger(__name__)


class CalingenConfig(AppConfig):
    """Application-specific configuration class, as required by Django.

    This sub-class of Django's ``AppConfig`` provides application-specific
    information to be used in Django's application registry (see
    :djangoapi:`Configuring Applications <applications/#configuring-applications>`).
    """

    name = "calingen"
    verbose_name = "CalInGen"

    def ready(self) -> None:
        """Apply app-specific stuff.

        Notes
        -----
        This method is executed when the application is (completely) loaded.

        It performs the following actions:

        - register the app's contributions to Django's
          :djangodoc:`System check framework <topics/checks/>` (see
          :mod:`calingen.checks` for implementation details);
        - inject app-specific settings into the project's ``settings`` module,
          if they are not already specified (see
          :mod:`calingen.settings` for details);
        """
        # delay app imports until now, to make sure everything else is ready
        # app imports
        from calingen import settings as app_default_settings
        from calingen.checks import (
            check_config_value_compiler,
            check_config_value_event_provider_notification,
            check_session_enabled,
        )

        # inject app-specific settings
        # see https://stackoverflow.com/a/47154840
        for name in dir(app_default_settings):
            if name.isupper() and not hasattr(settings, name):
                value = getattr(app_default_settings, name)
                logger.info("Injecting setting {} with value {}".format(name, value))
                setattr(settings, name, value)

        # register app-specific check functions
        register_check(check_config_value_compiler)
        register_check(check_config_value_event_provider_notification)
        register_check(check_session_enabled)



# SPDX-License-Identifier: MIT

"""The actual implementations of :class:`~calingen.interfaces.plugin_api.EventProvider`.

The classes just collect the pre-defined holidays in their ``entries``
class attribute.

The pre-defined holidays use :py:obj:`dateutil.rrule.rrule` to define their
recurrence. Most of Germany's holidays are based off the Easter date, so
``rrule`` allows for easy definitions.
"""

# Python imports
from datetime import datetime

# Django imports
from django.utils.translation import gettext_lazy as _

# external imports
from dateutil.rrule import WE, YEARLY, rrule

# app imports
from calingen.constants import EventCategory
from calingen.interfaces.plugin_api import EventProvider

# The following constants simply provide all the available German Holidays
# These will be combined in the actual implementation classes, e.g. in
# GermanyFederal
NEUJAHR = (
    _("New Year"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1)),
)
"""New Year, January, 1st."""

HEILIGE_DREI_KOENIGE = (
    _("Epiphany"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 6)),
)
"""Epiphany, January, 6th."""

FRAUENTAG = (
    _("Women's Day"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 3, 8)),
)
"""Women's Day, March, 8th."""

KARFREITAG = (
    _("Good Friday"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=-2),
)
"""Good Friday, the Friday before Easter.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

OSTER_SONNTAG = (
    _("Easter Sunday"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=0),
)
"""Easter Sunday.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

OSTER_MONTAG = (
    _("Easter Monday"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=1),
)
"""Easter Monday, the Monday after Easter.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

TAG_DER_ARBEIT = (
    _("Worker's Day"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 5, 1)),
)
"""Worker's Day, May, 1st."""

CHRISTI_HIMMELFAHRT = (
    _("Ascension"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=39),
)
"""Ascension, the Thursday 39 days after Easter.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

PFINGST_SONNTAG = (
    _("Pentecost Sunday"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=49),
)
"""Pentecost Sunday, the Sunday 49 days after Easter.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

PFINGST_MONTAG = (
    _("Pentecost Monday"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=50),
)
"""Pentecost Monday, the Monday 50 days after Easter.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

FRONLEICHNAM = (
    _("Corpus Christi"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 1, 1), byeaster=60),
)
"""Corpus Christi, the Thursday 60 days after Easter.

Notes
-----
This holiday is based off the Easter date, which shifts from year to year.
"""

MARIA_HIMMELFAHRT = (
    _("Assumption of Mary"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 8, 15)),
)
"""Assumption of Mary, August, 15th."""

WELTKINDERTAG = (
    _("Children's Day"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 9, 20)),
)
"""Children's Day, September, 20th."""

TAG_DER_DEUTSCHEN_EINHEIT = (
    _("Day of German Unity"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 10, 3)),
)
"""Day of German Unity, October, 3rd."""

REFORMATIONSTAG = (
    _("Reformation Day"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 10, 31)),
)
"""Reformation Day, October, 31st."""

ALLERHEILIGEN = (
    _("All Hallows"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 11, 1)),
)
"""All Hallows, November, 1st."""

BUSS_UND_BETTAG = (
    _("Day of Repetance"),
    EventCategory.HOLIDAY,
    rrule(
        freq=YEARLY,
        dtstart=datetime(1990, 1, 1),
        bymonth=11,
        byweekday=WE,
        bymonthday=(16, 17, 18, 19, 20, 21, 22),
    ),
)
"""Day of Repetance, the last Wednesday of November."""

ERSTER_WEIHNACHTSTAG = (
    _("Christmas Day"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 12, 25)),
)
"""Christmas Day, December, 25th."""

ZWEITER_WEIHNACHTSTAG = (
    _("Boxing Day"),
    EventCategory.HOLIDAY,
    rrule(freq=YEARLY, dtstart=datetime(1990, 12, 26)),
)
"""Boxing Day, December, 26th."""

# The actual classes will use Germany's federal structure as a taxononmy.
# However, there are some holidays that are applicable to all states, these are
# aggregated in this list.
# Please note: OSTER_SONNTAG and PFINGS_SONNTAG are not really considered
# "holiday" in all states, but they are sundays nonetheless.
FEDERAL_HOLIDAYS = [
    NEUJAHR,
    KARFREITAG,
    OSTER_SONNTAG,
    OSTER_MONTAG,
    TAG_DER_ARBEIT,
    CHRISTI_HIMMELFAHRT,
    PFINGST_SONNTAG,
    PFINGST_MONTAG,
    TAG_DER_DEUTSCHEN_EINHEIT,
    ERSTER_WEIHNACHTSTAG,
    ZWEITER_WEIHNACHTSTAG,
]


class GermanyFederal(EventProvider):
    """Provides federal holidays of Germany."""

    title = _("German Federal Holidays")

    entries = FEDERAL_HOLIDAYS


class BadenWuerttemberg(GermanyFederal):
    """Provides holidays of Baden-Wuerttemberg."""

    title = _("Holidays of Baden-Württemberg")

    entries = FEDERAL_HOLIDAYS + [HEILIGE_DREI_KOENIGE, FRONLEICHNAM, ALLERHEILIGEN]


class Bayern(GermanyFederal):
    """Provides holidays of Bayern."""

    title = _("Holidays of Bayern")

    entries = FEDERAL_HOLIDAYS + [
        HEILIGE_DREI_KOENIGE,
        FRONLEICHNAM,
        MARIA_HIMMELFAHRT,
        ALLERHEILIGEN,
    ]


class Berlin(GermanyFederal):
    """Provides holidays of Berlin."""

    title = _("Holidays of Berlin")

    entries = FEDERAL_HOLIDAYS + [FRAUENTAG]


class Brandenburg(GermanyFederal):
    """Provides holidays of Brandenburg."""

    title = _("Holidays of Brandenburg")

    entries = FEDERAL_HOLIDAYS + [REFORMATIONSTAG]


class Bremen(Brandenburg):
    """Provides holidays of Bremen."""

    title = _("Holidays of Bremen")


class Hamburg(Brandenburg):
    """Provides holidays of Hamburg."""

    title = _("Holidays of Hamburg")


class MecklenburgVorpommern(Brandenburg):
    """Provides holidays of Mecklenburg-Vorpommern."""

    title = _("Holidays of Mecklenburg-Vorpommern")


class Niedersachsen(Brandenburg):
    """Provides holidays of Niedersachsen."""

    title = _("Holidays of Niedersachsen")


class SchleswigHolstein(Brandenburg):
    """Provides holidays of Schleswig-Holstein."""

    title = _("Holidays of Schleswig-Holstein")


class Hessen(GermanyFederal):
    """Provides holidays of Hessen."""

    title = _("Holidays of Hessen")

    entries = FEDERAL_HOLIDAYS + [FRONLEICHNAM]


class NordrheinWestphalen(GermanyFederal):
    """Provides holidays of Nordrhein-Westphalen."""

    title = _("Holidays of Nordrhein-Westphalen")

    entries = FEDERAL_HOLIDAYS + [FRONLEICHNAM, ALLERHEILIGEN]


class RheinlandPfalz(NordrheinWestphalen):
    """Provides holidays of Rheinland-Pfalz."""

    title = _("Holidays of Rheinland-Pfalz")


class Saarland(GermanyFederal):
    """Provides holidays of Saarland."""

    title = _("Holidays of Saarland")

    entries = FEDERAL_HOLIDAYS + [FRONLEICHNAM, ALLERHEILIGEN, MARIA_HIMMELFAHRT]


class Sachsen(GermanyFederal):
    """Provides holidays of Sachsen."""

    title = _("Holidays of Sachsen")

    entries = FEDERAL_HOLIDAYS + [FRONLEICHNAM, BUSS_UND_BETTAG]


class SachsenAnhalt(GermanyFederal):
    """Provides holidays of Sachsen-Anhalt."""

    title = _("Holidays of Sachsen-Anhalt")

    entries = FEDERAL_HOLIDAYS + [HEILIGE_DREI_KOENIGE, REFORMATIONSTAG]


class Thueringen(GermanyFederal):
    """Provides holidays of Thueringen."""

    title = _("Holidays of Thüringen")

    entries = FEDERAL_HOLIDAYS + [FRONLEICHNAM, WELTKINDERTAG, REFORMATIONSTAG]



# SPDX-License-Identifier: MIT

"""App-specific mixins to be used with class-based views."""

# Django imports
from django.core.exceptions import ImproperlyConfigured

# app imports
from calingen.interfaces.data_exchange import CalendarEntryList
from calingen.models.event import Event
from calingen.models.profile import Profile


class RestrictToUserMixin:
    """Limits the resulting queryset to objects, that belong to the current user.

    This mixin overwrites the view's ``get_queryset()`` method and automatically
    uses the model's app-specific ``calingen_manager`` with its
    ``filter_by_user()`` method.

    The view, that uses this mixin, must provide the attribute ``model`` to
    actually make this mixin usable by any class-based view of the app.
    """

    def get_queryset(self):
        """Return a queryset that only contains objects, that belong to the current user.

        Internally, the returned queryset will use the app- and model-specific
        implementation of :class:`django.db.models.Manager` as provided by
        ``calingen_manager`` on all of the app's models. All of these managers
        rely on an app- and model-specific implementation of
        :class:`django.db.models.QuerySet` that provide a model-specific method
        ``filter_by_user()``.

        Returns
        -------
        django.db.models.QuerySet
            The actual returned QuerySet is provided by the model's app-specific
            manager ``calingen_manager``.

        Raises
        ------
        django.core.exceptions.ImproperlyConfigured
            The attribute ``model`` has to be defined on the view.
        """
        if self.model is None:
            raise ImproperlyConfigured(
                "{} is missing the 'model' attribute!".format(self.__class__.__name__)
            )

        return self.model.calingen_manager.get_queryset().filter_by_user(
            self.request.user
        )


class ProfileIDMixin:
    """Injects the :class:`~calingen.models.profile.Profile` id of the ``request.user`` into the context.

    This mixin uses ``get_context_data()`` to inject the ``profile_id``.

    The mixin can safely be used on any ``View`` that also uses the
    :class:`django.contrib.auth.mixins.LoginRequiredMixin`.
    """

    def get_context_data(self, **kwargs):  # noqa: D102
        context = super().get_context_data(**kwargs)

        context["profile_id"] = Profile.calingen_manager.get_profile(
            self.request.user
        ).id

        return context


class AllCalendarEntriesMixin:
    """Add all :class:`~calingen.interfaces.data_exchange.CalendarEntry` of the ``request.user``.

    This mixin uses ``get_context_data()`` to provide all of the user's
    :class:`~calingen.interfaces.data_exchange.CalendarEntry` instances for
    the context.
    """

    def get_context_data(self, **kwargs):  # noqa: D102
        context = super().get_context_data(**kwargs)

        # get the user's profile (required to process plugins)
        profile = Profile.calingen_manager.get_profile(self.request.user)

        all_entries = CalendarEntryList()
        internal_events = Event.calingen_manager.get_calendar_entry_list(
            user=self.request.user, year=context["target_year"]
        )
        plugin_events = profile.resolve(year=context["target_year"])
        all_entries.merge(internal_events)
        all_entries.merge(plugin_events)
        entries = all_entries.sorted()

        context["entries"] = entries

        # Usually, CalingenUserProfileIDMixin provides the (required)
        # profile_id, but as this view fetches the Profile anyway, it will be
        # added manually
        context["profile_id"] = profile.id

        return context



# SPDX-License-Identifier: MIT

"""The actual implementation of the compiler."""


# Django imports
from django.shortcuts import render

# app imports
from calingen.interfaces.plugin_api import CompilerProvider


class CopyPasteCompiler(CompilerProvider):
    """Do not compile sources, but just return them in a view for copy/paste."""

    title = "CopyPasteCompiler"

    @classmethod
    def get_response(cls, source, *args, **kwargs):  # noqa: D102
        context = {"rendered_source": source}
        return render(
            None,
            template_name="calingen/contrib/compiler_copypaste.html",
            context=context,
        )  # pragma: nocover



# SPDX-License-Identifier: MIT

"""Views related to the :class:`calingen.models.event.Event` model."""

# Django imports
from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse_lazy
from django.views import generic

# app imports
from calingen.models.event import Event, EventForm
from calingen.models.profile import Profile
from calingen.views.mixins import ProfileIDMixin, RestrictToUserMixin


class EventCreateView(LoginRequiredMixin, ProfileIDMixin, generic.CreateView):
    """Generic class-based view implementation to add :class:`calingen.models.event.Event` objects.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.CreateView`.

    The view relies on :class:`calingen.models.event.EventForm` as its
    ``ModelForm``. This is required to make custom validation logic available.
    """

    model = Event
    """Required attribute to tie this view to the model."""

    form_class = EventForm
    """Specify which form to use."""

    template_name_suffix = "_create"
    """Make the view use the template ``calingen/event_create.html``."""

    success_url = reverse_lazy("calingen:event-list")
    """The URL to redirect to after successfully creating the instance."""

    def form_valid(self, form):  # pragma: nocover
        """Inject the user's :class:`~calingen.models.profile.Profile` into the form.

        Every user does have only one :class:`~calingen.models.profile.Profile`
        and :class:`~calingen.models.event.Event` instances are directly tied
        to that ``Profile``.

        The :class:`~calingen.models.event.EventForm`` does not include the
        field to select the ``Profile``, as user's are only allowed to create
        ``Event`` instances for themselves.

        This method `injects` the user's ``Profile`` into the ``form.instance``
        to maintain database integrity.
        """
        # get the user's Profile to inject this into the instance
        form.instance.profile = Profile.objects.get(owner=self.request.user)

        return super().form_valid(form)


class EventDeleteView(
    RestrictToUserMixin,
    LoginRequiredMixin,
    ProfileIDMixin,
    generic.DeleteView,
):
    """Generic class-based view implementation to delete :class:`calingen.models.event.Event` objects.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.DeleteView`.
    """

    model = Event
    """Required attribute to tie this view to the model."""

    context_object_name = "event_item"
    """Provide a semantic name for the built-in context."""

    pk_url_kwarg = "event_id"
    """The name of the keyword argument as provided in the app's url configuration.

    By default, this is simply ``"pk"``, but for clarity, the app's url
    configuration (:mod:`calingen.urls`) uses the more explicit ``"event_id"``.
    """

    success_url = reverse_lazy("calingen:event-list")
    """The URL to redirect to after successfully deleting the instance."""


class EventDetailView(
    RestrictToUserMixin,
    LoginRequiredMixin,
    ProfileIDMixin,
    generic.DetailView,
):
    """Provide details of a :class:`calingen.models.event.Event` instance.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.DetailView`.
    """

    model = Event
    """Required attribute to tie this view to the model."""

    context_object_name = "event_item"
    """Provide a semantic name for the built-in context."""

    pk_url_kwarg = "event_id"
    """The name of the keyword argument as provided in the app's url configuration.

    By default, this is simply ``"pk"``, but for clarity, the app's url
    configuration (:mod:`calingen.urls`) uses the more explicit ``"event_id"``.
    """


class EventListView(
    RestrictToUserMixin,
    LoginRequiredMixin,
    ProfileIDMixin,
    generic.ListView,
):
    """Provide a list of :class:`calingen.models.event.Event` instances.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.ListView`.
    """

    model = Event
    """Required attribute to tie this view to the model."""

    context_object_name = "event_list"
    """Provide a semantic name for the built-in context."""


class EventUpdateView(
    RestrictToUserMixin,
    LoginRequiredMixin,
    ProfileIDMixin,
    generic.UpdateView,
):
    """Generic class-based view implementation to update :class:`calingen.models.event.Event` objects.

    Notes
    -----
    This implementation uses Django's generic class-based view
    :class:`django.views.generic.UpdateView`.

    The view relies on :class:`calingen.models.event.EventForm` as its
    ``ModelForm``. This is required to make custom validation logic available.
    """

    model = Event
    """Required attribute to tie this view to the model."""

    form_class = EventForm
    """Specify which form to use."""

    pk_url_kwarg = "event_id"
    """The name of the keyword argument as provided in the app's url configuration.

    By default, this is simply ``"pk"``, but for clarity, the app's url
    configuration (:mod:`calingen.urls`) uses the more explicit ``"event_id"``.
    """

    template_name_suffix = "_update"
    """Make the view use the template ``calingen/event_update.html``."""

    success_url = reverse_lazy("calingen:event-list")
    """The URL to redirect to after successfully updating the instance."""
